class BuildInfo:  # pylint: disable=too-few-public-methods
    python_version = '3.10.15'
    os_name = 'posix:ubuntu'
    version = '1.0.8'
    git_sha = '80fa8dcd6c2f7acc77c515f2b74affa646b26e03'
    git_branch = 'master'
    git_uncommitted = [
        'M  iuv/test_iuv_logger.py',
        'M  iuv/test_iuv_report.py',
        'D  medver_pytest/__init__.py',
        'D  medver_pytest/lib/.gitignore',
        'R  medver_pytest/lib/__init__.py -> src/__init__.py',
        'A  src/medver_pytest/.gitignore',
        'A  src/medver_pytest/__init__.py',
        'AM src/medver_pytest/build_info.py',
        'R  medver_pytest/lib/cfg.py -> src/medver_pytest/cfg.py',
        'R  medver_pytest/lib/cli.py -> src/medver_pytest/cli.py',
        'R  medver_pytest/lib/constants.py -> src/medver_pytest/constants.py',
        'RM medver_pytest/lib/constants_version.py -> src/medver_pytest/constants_version.py',
        'R  medver_pytest/lib/harness.py -> src/medver_pytest/harness.py',
        'R  medver_pytest/lib/iuv/__init__.py -> src/medver_pytest/iuv/__init__.py',
        'R  medver_pytest/lib/iuv/iuv.py -> src/medver_pytest/iuv/iuv.py',
        'R  medver_pytest/lib/iuv/iuv_data.py -> src/medver_pytest/iuv/iuv_data.py',
        'R  medver_pytest/lib/iuv/iuv_requirements.py -> src/medver_pytest/iuv/iuv_requirements.py',
        'R  medver_pytest/lib/report/__init__.py -> src/medver_pytest/log/__init__.py',
        'R  medver_pytest/lib/log/logger.py -> src/medver_pytest/log/logger.py',
        'R  medver_pytest/lib/log/logger_stdout.py -> src/medver_pytest/log/logger_stdout.py',
        'R  medver_pytest/lib/os_specific.py -> src/medver_pytest/os_specific.py',
        'R  medver_pytest/lib/protocol.py -> src/medver_pytest/protocol.py',
        'R  medver_pytest/lib/report/common/__init__.py -> src/medver_pytest/report/__init__.py',
        'R  medver_pytest/lib/report/protocol/__init__.py -> src/medver_pytest/report/common/__init__.py',
        'R  medver_pytest/lib/report/common/footer.py -> src/medver_pytest/report/common/footer.py',
        'R  medver_pytest/lib/report/common/gen_base_docx.py -> src/medver_pytest/report/common/gen_base_docx.py',
        'R  medver_pytest/lib/report/common/gen_base_minhtml.py -> src/medver_pytest/report/common/gen_base_minhtml.py',
        'R  medver_pytest/lib/report/common/gen_base_pdf.py -> src/medver_pytest/report/common/gen_base_pdf.py',
        'R  medver_pytest/lib/report/common/gen_base_txt.py -> src/medver_pytest/report/common/gen_base_txt.py',
        'R  medver_pytest/lib/report/common/header.py -> src/medver_pytest/report/common/header.py',
        'R  medver_pytest/lib/report/common/page_details.py -> src/medver_pytest/report/common/page_details.py',
        'R  medver_pytest/lib/report/common/page_info.py -> src/medver_pytest/report/common/page_info.py',
        'R  medver_pytest/lib/report/summary/__init__.py -> src/medver_pytest/report/protocol/__init__.py',
        'R  medver_pytest/lib/report/protocol/gen_tp_base.py -> src/medver_pytest/report/protocol/gen_tp_base.py',
        'R  medver_pytest/lib/report/protocol/gen_tp_docx.py -> src/medver_pytest/report/protocol/gen_tp_docx.py',
        'R  medver_pytest/lib/report/protocol/gen_tp_minhtml.py -> src/medver_pytest/report/protocol/gen_tp_minhtml.py',
        'R  medver_pytest/lib/report/protocol/gen_tp_pdf.py -> src/medver_pytest/report/protocol/gen_tp_pdf.py',
        'R  medver_pytest/lib/report/protocol/gen_tp_txt.py -> src/medver_pytest/report/protocol/gen_tp_txt.py',
        'R  medver_pytest/lib/report/report.py -> src/medver_pytest/report/report.py',
        'R  medver_pytest/lib/report/trace_matrix/__init__.py -> src/medver_pytest/report/summary/__init__.py',
        'R  medver_pytest/lib/report/summary/gen_summary_base.py -> src/medver_pytest/report/summary/gen_summary_base.py',
        'R  medver_pytest/lib/report/summary/gen_summary_docx.py -> src/medver_pytest/report/summary/gen_summary_docx.py',
        'R  medver_pytest/lib/report/summary/gen_summary_minhtml.py -> src/medver_pytest/report/summary/gen_summary_minhtml.py',
        'R  medver_pytest/lib/report/summary/gen_summary_pdf.py -> src/medver_pytest/report/summary/gen_summary_pdf.py',
        'R  medver_pytest/lib/report/summary/gen_summary_txt.py -> src/medver_pytest/report/summary/gen_summary_txt.py',
        'R  medver_pytest/lib/view/__init__.py -> src/medver_pytest/report/trace_matrix/__init__.py',
        'R  medver_pytest/lib/report/trace_matrix/gen_trace_matrix_base.py -> src/medver_pytest/report/trace_matrix/gen_trace_matrix_base.py',
        'R  medver_pytest/lib/report/trace_matrix/gen_trace_matrix_docx.py -> src/medver_pytest/report/trace_matrix/gen_trace_matrix_docx.py',
        'R  medver_pytest/lib/report/trace_matrix/gen_trace_matrix_minhtml.py -> src/medver_pytest/report/trace_matrix/gen_trace_matrix_minhtml.py',
        'R  medver_pytest/lib/report/trace_matrix/gen_trace_matrix_pdf.py -> src/medver_pytest/report/trace_matrix/gen_trace_matrix_pdf.py',
        'R  medver_pytest/lib/report/trace_matrix/gen_trace_matrix_txt.py -> src/medver_pytest/report/trace_matrix/gen_trace_matrix_txt.py',
        'R  medver_pytest/lib/requirements.py -> src/medver_pytest/requirements.py',
        'R  medver_pytest/lib/result_formatter.py -> src/medver_pytest/result_formatter.py',
        'R  medver_pytest/lib/result_summary.py -> src/medver_pytest/result_summary.py',
        'R  medver_pytest/lib/services.py -> src/medver_pytest/services.py',
        'R  medver_pytest/lib/status_info.py -> src/medver_pytest/status_info.py',
        'R  medver_pytest/lib/stdout_logger.py -> src/medver_pytest/stdout_logger.py',
        'A  src/medver_pytest/storage/__init__.py',
        'R  medver_pytest/lib/storage/storage.py -> src/medver_pytest/storage/storage.py',
        'R  medver_pytest/lib/storage/storage_dev.py -> src/medver_pytest/storage/storage_dev.py',
        'R  medver_pytest/lib/storage/storage_file_base.py -> src/medver_pytest/storage/storage_file_base.py',
        'R  medver_pytest/lib/storage/storage_shared_drive.py -> src/medver_pytest/storage/storage_shared_drive.py',
        'R  medver_pytest/lib/summary.py -> src/medver_pytest/summary.py',
        'R  medver_pytest/lib/trace_matrix.py -> src/medver_pytest/trace_matrix.py',
        'R  medver_pytest/lib/utils.py -> src/medver_pytest/utils.py',
        'R  medver_pytest/lib/verifier.py -> src/medver_pytest/verifier.py',
        'A  src/medver_pytest/view/__init__.py',
        'R  medver_pytest/lib/view/dlg_base.py -> src/medver_pytest/view/dlg_base.py',
        'R  medver_pytest/lib/view/dlg_manual_action.py -> src/medver_pytest/view/dlg_manual_action.py',
        'R  medver_pytest/lib/view/dlg_manual_verify.py -> src/medver_pytest/view/dlg_manual_verify.py',
        'R  medver_pytest/lib/view/view.py -> src/medver_pytest/view/view.py',
        'MM tools/pyproject_template.toml',
        ' m tools/xplat_utils',
        'M  ut/mocks/mock_protocol.py',
        'M  ut/test_ut_cfg.py',
        'M  ut/test_ut_cli.py',
        'M  ut/test_ut_harness.py',
        'M  ut/test_ut_page_details.py',
        'M  ut/test_ut_requirements.py',
        'M  ut/test_ut_result_summary.py',
        'M  ut/test_ut_storage.py',
        'M  ut/test_ut_utils.py',
        'M  ut/test_ut_verify_delta.py',
        'M  ut/test_ut_verify_delta_pct.py',
        'M  xplat.cfg',
    ]
    git_unpushed = [
    ]



import datetime
import os
import sys

from medver_pytest import services


# -------------------
## holds various utility functions
class Utils:

    # -------------------
    ## gets the location of the caller
    #
    # @param levels   default 3; used to go back further in the trace back
    # @return a string showing the filename and line number of the caller
    @staticmethod
    def get_location(levels=3):
        # get a full stackframe
        tb = None
        try:
            raise AssertionError
        except AssertionError:
            tb = sys.exc_info()[2]

        # go levels callers back
        frame = tb.tb_frame
        while levels > 0:
            frame = frame.f_back
            levels -= 1

        # uncomment to debug
        # logger.line('')
        # logger.dbg(frame.f_code.co_name)
        # logger.dbg(frame.f_code.co_filename)
        # logger.dbg(frame.f_lineno)

        fname = os.path.basename(frame.f_code.co_filename)
        location = f'{fname}({frame.f_lineno})'
        return location

    # -------------------
    ## gets the current time in utc or local time
    #
    # @param use_cfg_fmt  if True, use the format given in cfg.json, otherwise use short ofrm
    # @param use_time     if None, use the current UTC, otherwise use the given time
    # @return a string showing the current date_time with
    @staticmethod
    def get_dts(use_cfg_fmt, use_time=None):
        # assume use_time is UTC
        if use_time is not None:
            dts = use_time
        else:
            dts = Utils.get_utc()

        if not services.cfg.use_utc:
            dts = dts.astimezone()

        if use_cfg_fmt:
            dts = dts.strftime(services.cfg.dts_format)
        else:
            dts = dts.strftime('%Y-%m-%d %H:%M:%S')
        return dts

    # -------------------
    ## gets the current UTC time as a datetime
    #
    # @return a datetime object with UTC time in it
    @staticmethod
    def get_utc():
        return datetime.datetime.now(datetime.timezone.utc)



from .storage_file_base import StorageFileBase


# -------------------
## Holds storage information and how/where to save it
class StorageDev(StorageFileBase):
    # -------------------
    ## constructor
    def __init__(self):
        super().__init__()

    # -------------------
    ## initialize
    #
    # @return None
    def init(self):
        pass

    # -------------------
    ## terminate
    #
    # @return None
    def term(self):
        pass



# -------------------
## Holds a result summary
class ResultSummary(dict):
    # -------------------
    ## constructor
    def __init__(self):
        super().__init__()
        ## the result so far: PASS/FAIL
        self.result = 'PASS'

        ## the raw actual value
        self.actual = None
        ## the original python type of the actual value
        self.actual_type = None
        ## the formatted string for display in the report of the actual value
        self.actual_formatted = None

        ## the raw expected value
        self.expected = None
        ## the original python type of the expected value
        self.expected_type = None
        ## the formatted string for display in the report of the expected value
        self.expected_formatted = None

        ## the list of reqids (if any) for this result
        self.reqids = None

        ## the verify_xx() location that generated this result
        self.location = None

        ## the formatter supplied by the user, if any
        self.formatter = None

    # -------------------
    ## load from a json object
    #
    # @param j   the json object
    # @return None
    def append_result(self, j):
        # overwrite these attributes
        self.result = j['result']
        self.actual = j['actual']
        self.actual_formatted = j['actual_formatted']
        self.actual_type = j['actual_type']
        self.expected = j['expected']
        self.expected_formatted = j['expected_formatted']
        self.expected_type = j['expected_type']
        self.location = j['location']
        self.formatter = j['formatter']

        # add on to the list of reqids
        if self.reqids is None:
            self.reqids = {}

        if j['reqids'] is None:
            pass
        else:
            for reqid in j['reqids'].keys():
                self.reqids[reqid] = 1

    # -------------------
    ## indicate this result has passed
    #
    # @return None
    def passed(self):
        self.result = 'PASS'

    # -------------------
    ## indicate this result has failed
    #
    # @return None
    def failed(self):
        self.result = 'FAIL'

    # -------------------
    ## set the reqids to the given list or string
    # Note: requirement ids are converted to upper case
    #
    # @param reqids   either None, a single string, a list of strings
    # @return None
    def set_reqids(self, reqids):
        if reqids is None:
            self.reqids = None
        elif isinstance(reqids, str):
            self.reqids = {reqids.upper(): 1}
        elif isinstance(reqids, list):
            self.reqids = {}
            for reqid in reqids:
                self.reqids[reqid.upper()] = 1

    # -------------------
    ## required for JSON to handle this class as a dictionary
    #
    # @param key      the name of the attribute to set
    # @param value    the value of the attribute
    # @return None
    def __setattr__(self, key, value):
        self[key] = value

    # -------------------
    ## required for JSON to handle this class as a dictionary
    #
    # @param key      the name of the attribute to get
    # @return the value of the named attribute
    def __getattr__(self, key):
        return self[key]

    # -------------------
    ## generate a formatted version of the actual value
    #
    # @param val    the value to format
    # @return None
    def format_actual(self, val):
        self.actual_type = type(val).__name__
        self.actual = str(val)
        if self.formatter is not None and self.formatter.actual_tag is not None:
            # override actual value with the tag the user requested
            self.actual_formatted = self.formatter.actual_tag
        else:
            formatted_val = self._format('format_actual', val, self.actual_type)
            self.actual_formatted = formatted_val

    # -------------------
    ## generate a formatted version of the expected value
    #
    # @param val      the value to format
    # @param prefix   (optional) string; a prefix on the formatted string
    # @param suffix   (optional) string; a suffix on the formatted string
    # @param is_not   (optional) bool; if True, add "not" as a prefix on the formatted string
    # @return None
    def format_expected(self, val, prefix=None, suffix=None, is_not=False):
        self.expected_type = type(val).__name__
        self.expected = str(val)
        if prefix is None:
            prefix = ''

        if suffix is None:
            suffix = ''

        if is_not:
            prefix = f'not {prefix}'

        formatted_val = self._format('format_expected', val, self.expected_type)
        self.expected_formatted = f'{prefix}{formatted_val}{suffix}'

    # -------------------
    ## generate a formatted version of the given regex
    #
    # @param val      the regex to format
    # @param is_not   (optional) bool; if True, add "not" as a prefix on the formatted string
    # @return None
    def format_regex_expected(self, val, is_not=False):
        self.expected_type = type(val).__name__
        self.expected = val
        if is_not:
            prefix = 'does not match'
        else:
            prefix = 'matches'
        self.expected_formatted = f'{prefix} r\'{str(val)}\''

    # -------------------
    ## generate a formatted version of the given delta
    #
    # @param val      the expected value to format
    # @param delta    numeric; the +/- value to use
    # @param is_pct   (optional) bool; if True, delta is percentage
    # @param is_not   (optional) bool; if True, add "outside" as a prefix on the formatted string
    # @return None
    def format_delta_expected(self, val, delta, is_pct=False, is_not=False):
        self.expected_type = type(val).__name__
        self.expected = val
        if is_not:
            prefix = 'outside '
        else:
            prefix = ''
        if is_pct:
            suffix = '%'
        else:
            suffix = ''
        self.expected_formatted = f'{prefix}{val} +/-{delta}{suffix}'

    # -------------------
    ## generate a formatted version of the given value
    #  no formatter: convert to string
    #  format given: generate string using the given format
    #  as_hex      : generate string using default hex format
    #
    # @param tag             used for logging purposes
    # @param val             the value to format
    # @param original_type   the python object type of the value
    # @return if successful a formatted string, otherwise None
    def _format(self, tag, val, original_type):
        if self.formatter is None:
            # no formatter, so just convert to a simple string
            return str(val)

        if self.formatter.format is not None:
            # user gave a format, so format it with their request
            return self.formatter.format % val

        if self.formatter.as_hex:
            return self._format_as_hex(tag, val, original_type)

        # formatter was supplied, but nothing to do,
        # so just convert to a simple string
        return str(val)

    # -------------------
    ## generate a formatted version of the given value as hex
    #   * int       : format as an 8 byte hex string
    #   * str       : convert to hex bytes
    #   * bytes     : format as a matrix of hex chars (as a string)
    #   * bytearray : format as a matrix of hex chars (as a string)
    #
    # @param tag             used for logging purposes
    # @param val             the value to format
    # @param original_type   the python object type of the value
    # @return if successful a formatted string, otherwise None
    def _format_as_hex(self, tag, val, original_type):  # pylint: disable=unused-argument
        if original_type == 'int':
            return f'0x{val:08X}'

        if original_type == 'str':
            val = bytes(val, "utf-8")

        if original_type in ['bytes', 'bytearray', 'str']:
            msg = ''
            count = 0
            for ch in val:
                count += 1
                if ch == val[-1]:
                    suffix = ''
                elif count == 8:
                    suffix = '\n'
                    count = 0
                else:
                    suffix = ' '
                msg += f'{ch:02X}{suffix}'
            return msg

        # nothing to do for this type,
        return str(val)



import tkinter as tk
import tkinter.scrolledtext as tkst

from .dlg_base import DlgBase
from medver_pytest import services


# -------------------
## holds Dialog Box for a Manual Action Step
class DlgManualAction(DlgBase):
    # -------------------
    ## constructor
    def __init__(self):
        ## see base class for doc
        self._dlg = None

        ## see base class for doc
        self._dlg = None
        super().__init__()

        ## the return state, holds abort, done, or cantdo
        self._return_value = 'initial'
        ## the reason text if state is cantdo, otherwise ''
        self._reason = ''
        ## whether reason was entered or not
        self._reason_entered = False

        # widgets
        ## the Done btn
        self._btn_done = None
        ## the Can't do btn
        self._btn_cantdo = None
        ## the Reason Text widget
        self._txt_reason = None

    # -------------------
    ## show dialog for a manual action
    #
    # @param parent   the tk root window
    # @param title    the dlg title
    # @param action   the action for the tester to perform
    # @return the final state of the dlg (done, cantdo, abort), the reason if state is cantdo, otherwise ''
    def step_manual_action(self, parent, title, action):
        ## see base class for doc
        self._parent = parent
        self._show_dlg_manual_action(title, action)
        self._parent.wait_window(self._dlg)

        return self._return_value, self._reason

    # -------------------
    ## show the dlg for performing a manual action
    #
    # @param title   the dlg title
    # @param action  the action for the tester to perform
    # @return None
    def _show_dlg_manual_action(self, title, action):
        row = 0
        msg = 'Press Can\'t do or perform the action and press Done'
        row = self._dlg_common_first(row, title, msg, self._ensure_initials)

        row += 1
        self._create_action_desc(row, action)
        services.logger.user(f'action: {action}')

        row += 1
        self._create_empty_row(row)

        row += 1
        self._create_done_btn(row)

        row += 1
        self._create_cantdo_btn(row, 0)

        row += 1
        self._create_cell_lbl('Reason cannot be done:', row, 0)
        row += 1
        self._create_reason(row, 0)

        ## set callbacks for click and ensure initials
        self._dlg_common_last(row,
                              self._handle_click,
                              self._ensure_initials)

    # -------------------
    ## the action for the tester to perform
    #
    # @param row          the row to place it in
    # @param action_desc  the tester action needed for the current protocol step
    # @return None
    def _create_action_desc(self, row, action_desc):
        lbl = tkst.ScrolledText(self._frame,
                                wrap=tk.WORD,  # wrap text on a word boundary
                                height=5,  # lines
                                width=self.col0_txt_width,
                                font=self.common_font,
                                fg='black',
                                bg='lightgrey',
                                highlightbackground='black', highlightthickness=2)
        lbl.insert(1.0, action_desc)
        lbl.configure(state='disabled')  # make it readonly
        lbl.grid(row=row, column=0, sticky='WE', columnspan=3)

    # -------------------
    ## btn indicates action was done
    #
    # @param row     the row to place it in
    # @return None
    def _create_done_btn(self, row):
        self._btn_done = tk.Button(self._frame, text='Done',
                                   command=lambda: self._handle_click('done'),
                                   width=self.col0_btn_width,
                                   font=self.common_font,
                                   fg='black',
                                   bg='lightgreen',
                                   disabledforeground='lightgrey',
                                   highlightbackground='black', highlightthickness=2)
        self._btn_done.grid(row=row, column=0, sticky='W')

    # -------------------
    ## disable the Done button
    #
    # @return None
    def _done_disable(self):
        self._btn_done.config(state='disabled', fg='black', highlightbackground='black', highlightthickness=0)

    # -------------------
    ## enable the Done button
    #
    # @return None
    def _done_enable(self):
        self._btn_done.config(state='normal', fg='black', highlightbackground='black', highlightthickness=2)

    # -------------------
    ## place a highlight around the Done btn
    #
    # @return None
    def _done_highlight(self):
        self._btn_done.config(fg='black', highlightbackground='black', highlightthickness=2)

    # -------------------
    ## unhighlight the Done btn
    #
    # @return None
    def _done_unhighlight(self):
        self._btn_done.config(fg='black', highlightbackground='lightgrey', highlightthickness=2)

    # -------------------
    ## btn that indicates the tester can't do the requested action
    #
    # @param row     the row to place it in
    # @param col     the column to place it in
    # @return None
    def _create_cantdo_btn(self, row, col):
        self._btn_cantdo = tk.Button(self._frame, text='Can\'t Do',
                                     command=lambda: self._handle_click('cantdo'),
                                     width=self.col0_btn_width,
                                     font=self.common_font,
                                     fg='black',
                                     bg='gold',
                                     disabledforeground='lightgrey',
                                     highlightbackground='black', highlightthickness=2)
        self._btn_cantdo.grid(row=row, column=col, sticky='W')

    # -------------------
    ## disable the Can't Do button
    #
    # @return None
    def _cantdo_disable(self):
        self._btn_cantdo.config(state='disabled', fg='black', highlightbackground='black', highlightthickness=0)

    # -------------------
    ## enable the Can't Do button
    #
    # @return None
    def _cantdo_enable(self):
        self._btn_cantdo.config(state='normal', fg='black', highlightbackground='black', highlightthickness=2)

    # -------------------
    ## highlight the Can't Do btn
    #
    # @return None
    def _cantdo_highlight(self):
        self._btn_cantdo.config(fg='black', highlightbackground='black', highlightthickness=2)

    # -------------------
    ## unhighlight the Can't Do button
    #
    # @return None
    def _cantdo_unhighlight(self):
        self._btn_cantdo.config(fg='black', highlightbackground='lightgrey', highlightthickness=2)

    # -------------------
    ## Create the Reason text area. The tester fills
    # it in if they can not perform the requested action
    #
    # @param row     the row to place it in
    # @param col     the column to place it in
    # @return None
    def _create_reason(self, row, col):
        self._txt_reason = tkst.ScrolledText(self._frame,
                                             wrap=tk.WORD,  # wrap text on a word boundary
                                             height=5,  # lines
                                             width=self.col12_width,
                                             font=self.common_font,
                                             fg='black',
                                             bg='lightgrey',
                                             highlightbackground='black', highlightthickness=2)
        self._txt_reason.configure(state='disabled')  # initially, make it readonly
        self._txt_reason.focus()
        ## set callback for key release
        self._txt_reason.bind('<KeyRelease>', self._reason_key_pressed)
        self._txt_reason.grid(row=row, column=col, sticky='NEWS', columnspan=3)

    # -------------------
    ## handle an entry in the Reason box
    #
    # @param unused   ignored; the key that was pressed
    # @return None
    # noinspection PyUnusedLocal
    def _reason_key_pressed(self, unused):  # pylint: disable=unused-argument
        self._reason_entered = True
        self._get_reason()
        self._check_next()

    # -------------------
    ## get the content form the reason text box
    #
    # @return None
    def _get_reason(self):
        # get the reason text
        self._reason = self._txt_reason.get('1.0', 'end')
        self._reason = self._reason.strip()

    # -------------------
    ## check content of reason field;
    # should be 2 or more characters
    #
    # @return True if _reason has invalid content, False otherwise
    def _reason_is_wrong(self):
        return len(self._reason) < 2

    # -------------------
    ## disable the Reason text box
    #
    # @return None
    def _reason_disable(self):
        self._txt_reason['state'] = 'disabled'
        self._txt_reason['fg'] = 'darkgrey'
        self._txt_reason['bg'] = 'lightgrey'

    # -------------------
    ## enable the Reason text box
    # note: don't clear out reason text in case the tester wants to extend it
    #
    # @return None
    def _reason_enable(self):
        self._txt_reason['state'] = 'normal'
        self._txt_reason['fg'] = 'black'
        self._txt_reason['bg'] = 'white'

    # -------------------
    ## handle a dlg box click
    #
    # @param option  the btn the tester clicked
    # @return None
    def _handle_click(self, option):
        self._get_reason()
        if option == 'done':
            services.logger.user('click Done')
            self._reason_disable()
            self._cantdo_unhighlight()
            self._done_highlight()
            self._return_value = 'done'
            self._check_next()

        elif option == 'cantdo':
            services.logger.user('click Can\t Do')
            self._reason_enable()
            self._cantdo_highlight()
            self._done_unhighlight()
            self._return_value = 'cantdo'
            self._check_next()

        elif option == 'next':
            services.logger.user(f'reason: {self._reason}')
            services.logger.user('click Next')

            ok = self._check_next()
            if ok:
                self._dlg.destroy()

        elif option == 'abort':
            services.logger.user(f'reason: {self._reason}')
            services.logger.user('click Abort')
            self._return_value = 'abort'
            self._dlg.destroy()

    # -------------------
    ## check if Next button should be enabled or not.
    # Set message box as appropriate.
    #
    # @return True if all is ready, False otherwise
    def _check_next(self):
        ok = False
        if self._return_value == 'initial' and not self._reason_entered:
            self._next_disable()
            self._set_message('red', 'Perform step action, then Press Done or Press Can\'t Do')
        elif self._return_value == 'initial' and self._reason_entered:
            self._next_disable()
            # assume that if a value was entered in Reason, the action can't be done
            self._set_message('red', 'Check Reason, press Can\'t Do, and then press Next')
        elif self._return_value == 'cantdo' and self._reason_is_wrong():
            self._next_disable()
            self._set_message('red', 'Reason is too short. Enter full Reason.')
        elif self._return_value == 'cantdo' and not self._reason_is_wrong():
            self._next_enable()
            self._set_message('green', 'Can\'t Do selected. Check Reason, press Next when ready')
            ok = True
        elif self._return_value == 'done':
            self._next_enable()
            self._set_message('green', 'Done selected. Press Next when ready')
            ok = True
        else:
            services.logger.err(f'dlg_manual_action._check_next: invalid combination: {self._return_value} '
                                f'reason_entered:{self._reason_entered} '
                                f'reason_is_wrong:{self._reason_is_wrong()}')
            # TODO abort?
        return ok

    # -------------------
    ## disable buttons (except abort) if initials box is empty
    #
    # @return None
    def _ensure_initials(self):
        if self._tester_initials_is_wrong():
            self._reason_disable()
            self._done_disable()
            self._cantdo_disable()
            self._next_disable()
            self._set_message('red', 'Enter your initials (2 or 3 characters)')
        else:
            self._done_enable()
            self._cantdo_enable()
            self._check_next()



from .footer import Footer
from .header import Header
from .page_details import PageDetails
from medver_pytest import services


# -------------------
## holds page information for generating documents
class PageInfo:

    # -------------------
    ## constructor
    def __init__(self):
        ## the page orientation: landscape or portrait
        self.orientation = 'landscape'

        ## the page size: letter or a4
        self.page_size = 'letter'

        ## the header info
        self.header = Header()

        ## the footer info
        self.footer = Footer()

        ## the test report page details
        self._tp_report = PageDetails()
        ## default orientation for test report is landscapte
        self._tp_report.orientation = 'landscape'

        ## the test protocol results page details
        self._tp_protocol = PageDetails()
        ## default orientation for test protocol results is landscapte
        self._tp_protocol.orientation = 'landscape'

        ## the trace matrix page details
        self._trace = PageDetails()
        ## default orientation for trace matrix is landscapte
        self._trace.orientation = 'landscape'

        ## the summary page details
        self._summary = PageDetails()
        ## default orientation for trace matrix is portrait
        self._trace.orientation = 'portrait'

    # -------------------
    ## get a value from this object
    #
    # @param item   the name of the attribute to get
    # @return the value of the attribute
    def __getitem__(self, item):
        if item == 'header':
            return self.header

        if item == 'footer':
            return self.footer

        if item == 'orientation':
            return self.orientation

        if item == 'page_size':
            return self.page_size

        services.abort(f'bad item name: {item}')
        return None  # for lint

    # -------------------
    ## initialize all test protocol (no results) fields
    #
    # @param item  the value to set
    # @return None
    def init_tp_protocol(self, item):
        self._set_all_fields(item, self._tp_protocol)

    # -------------------
    ## initialize all test report (with results) with results fields
    #
    # @param item  the value to set
    # @return None
    def init_tp_report(self, item):
        self._set_all_fields(item, self._tp_report)

    # -------------------
    ## initialize all trace matrix fields
    #
    # @param item  the value to set
    # @return None
    def init_trace(self, item):
        self._set_all_fields(item, self._trace)

    # -------------------
    ## initialize all summary fields
    #
    # @param item  the value to set
    # @return None
    def init_summary(self, item):
        self._set_all_fields(item, self._summary)

    # -------------------
    ## initialize all test protocol (no results) cfg to default values
    #
    # @return None
    def set_tp_protocol_cfg(self):
        self._set_all_cfg(self._tp_protocol)

    # -------------------
    ## initialize all test report with results cfg to default values
    #
    # @return None
    def set_tp_report_cfg(self):
        self._set_all_cfg(self._tp_report)

    # -------------------
    ## initialize all trace matrix cfg to default values
    #
    # @return None
    def set_trace_cfg(self):
        self._set_all_cfg(self._trace)

    # -------------------
    ## initialize all summary cfg to default values
    #
    # @return None
    def set_summary_cfg(self):
        self._set_all_cfg(self._summary)

    # -------------------
    ## check some content
    #
    # @return number of errors found
    def check(self):
        errs = 0
        if self._tp_protocol.orientation != self._tp_report.orientation:
            errs += 1
            services.logger.err(f'tp_protocol orientation ({self._tp_protocol.orientation}) '
                                f'does not match tp_report ({self._tp_report.orientation})')

        return errs

    # -------------------
    ## generate a report to the given file
    #
    # @param fp  the file to write to
    # @return None
    def report(self, fp):
        fp.write(f"{'Header left': <20}: {self.header.left}\n")
        fp.write(f"{'Header middle': <20}: {self.header.middle}\n")
        fp.write(f"{'Header right': <20}: {self.header.right}\n")

        fp.write(f"{'Footer left': <20}: {self.footer.left}\n")
        fp.write(f"{'Footer middle': <20}: {self.footer.middle}\n")
        fp.write(f"{'Footer right': <20}: {self.footer.right}\n")

    # === Private

    # -------------------
    ## set the current values from the given source
    #
    # @param src   the source to use
    # @return None
    def _set_all_cfg(self, src):
        self.orientation = src.orientation
        self.page_size = src.page_size

        self.header.left = src.header.left
        self.header.middle = src.header.middle
        self.header.right = src.header.right

        self.footer.left = src.footer.left
        self.footer.middle = src.footer.middle
        self.footer.right = src.footer.right

    # -------------------
    ## set the various fields
    # * orientation
    # * page size
    # * headder: left, middle, right
    # * footer : left, middle, right
    #
    # @param src   the source to use
    # @param dst   the destinaton to save to
    # @return None
    def _set_all_fields(self, src, dst):
        self._set_field(src, dst, 'orientation')
        self._set_field(src, dst, 'page_size')
        self._set_sub_field(src, dst, 'header', 'left')
        self._set_sub_field(src, dst, 'header', 'middle')
        self._set_sub_field(src, dst, 'header', 'right')
        self._set_sub_field(src, dst, 'footer', 'left')
        self._set_sub_field(src, dst, 'footer', 'middle')
        self._set_sub_field(src, dst, 'footer', 'right')

    # -------------------
    ## set a destination field from the source
    #
    # @param src     the source to use
    # @param dst     the destinaton to save to
    # @param field   the field name
    # @return None
    def _set_field(self, src, dst, field):
        if field in src:
            dst[field] = src[field]

    # -------------------
    ## set a destination subfield from the source
    #
    # @param src     the source to use
    # @param dst     the destinaton to save to
    # @param parent  the parent field
    # @param field   the subfield name
    # @return None
    def _set_sub_field(self, src, dst, parent, field):
        if field in src[parent]:
            dst[parent][field] = src[parent][field]



from medver_pytest import services


# -------------------
## holds trace matrix data
class TraceMatrix:
    # -------------------
    ## constructor
    def __init__(self):
        ## holds the trace matrix info
        self._matrix = {}
        ## holds the current test script name
        self._current_test_script = None
        ## indicates if there is any information to be saved to the output file
        self._dirty = False

    # -------------------
    ## ensure that the current data is cleared out when the test script changes
    #
    # @return None
    def check_test_script(self):
        if self._current_test_script != services.cfg.test_script:
            if self._current_test_script is not None:
                self.save()

            self._current_test_script = services.cfg.test_script
            self._matrix = {}
            self._dirty = True

    # -------------------
    ## terminate
    #
    # @return None
    def term(self):
        self.save()

    # -------------------
    ## add protocol info to one or more reqids
    #
    # @param req_ids      one or more reqids to add
    # @param proto_id     the protocol id
    # @param proto_info   additional protocol information
    # @return None
    def add_proto(self, req_ids, proto_id, proto_info):
        self.check_test_script()

        for req_id in req_ids:
            if req_id in self._matrix:
                self._add_to(req_id, proto_id, proto_info)
            else:
                self._add(req_id, proto_id, proto_info)
        self.save()

    # -------------------
    ## add protocol info to a new req_id
    #
    # @param req_id       the reqid
    # @param proto_id     the protocol id
    # @param proto_info   additional protocol information
    # @return None
    def _add(self, req_id, proto_id, proto_info):
        self._matrix[req_id] = [
            {
                'proto_id': proto_id,
                'proto_info': proto_info,  # TODO what info?
            }
        ]
        self._dirty = True

    # -------------------
    ## add protocol info to an existing req_id
    #
    # @param req_id       the reqid
    # @param proto_id     the protocol id
    # @param proto_info   additional protocol information
    # @return None
    def _add_to(self, req_id, proto_id, proto_info):
        self._matrix[req_id].append(
            {
                'proto_id': proto_id,
                'proto_info': proto_info,  # TODO what info?
            }
        )
        self._dirty = True

    # -------------------
    ## save trace matrix data to storage
    #
    # @return None
    def save(self):
        services.storage.save_trace(self._matrix)
        self._dirty = False

    # === IUV only

    # -------------------
    ## (IUV only) get current dirty flag
    #
    # @return dirty flag
    def iuv_get_dirty(self):
        return self._dirty

    # -------------------
    ## (IUV only) get current trace object
    #
    # @return trace object
    def iuv_get_trace(self):
        return self._matrix



import json
import os

from medver_pytest import services


# --------------------
## holds all information that be used from run invocation to others
# Note:
#  * should hold only values that the user can/should change between runs
#  * no constants
class StatusInfo:
    # --------------------
    ## constructor
    def __init__(self):
        ## holds last manual tester initials if any
        self._last_tester_initials = ''

        ## holds last position of dlg box, x coord
        self._last_position_x = None
        ## holds last position of dlg box, y coord
        self._last_position_y = None
        ## holds path to the status json file
        self._path = os.path.join(services.cfg.statusdir, 'status.json')

        self._load()

    # -------------------
    ## getter for the last dlg x position
    #
    # @return last dlg position for x coordinate
    @property
    def last_position_x(self):
        return self._last_position_x

    # -------------------
    ## getter for the last dlg y position
    #
    # @return last dlg position for y coordinate
    @property
    def last_position_y(self):
        return self._last_position_y

    # -------------------
    ## check if there is a last position saved
    #
    # @return None
    def last_position_saved(self):
        return self._last_position_x is not None and self._last_position_y is not None

    # -------------------
    ## save last dlg screen coordinates
    #
    # @param x   the x coordinate of the screen position
    # @param y   the y coordinate of the screen position
    # @return None
    def save_last_position(self, x, y):
        self._last_position_x = x
        self._last_position_y = y
        self._save()

    # -------------------
    ## getter for the last tester initials
    #
    # @return last tester initials
    @property
    def last_tester_initials(self):
        return self._last_tester_initials

    # -------------------
    ## setter for the last tester initials
    #
    # @param val  the new last tester initials
    # @return None
    @last_tester_initials.setter
    def last_tester_initials(self, val):
        self._last_tester_initials = val
        self._save()

    # -------------------
    ## save all data to file
    #
    # @return None
    def _save(self):
        with open(self._path, 'w', encoding='utf-8') as fp:
            j = {
                'last_tester_initials': self._last_tester_initials,
                'last_position_x': self._last_position_x,
                'last_position_y': self._last_position_y,
            }
            json.dump(j, fp, indent=4)

    # -------------------
    ## save data from file, if it exists.
    # Otherewise leave data in its current state.
    #
    # @return None
    def _load(self):
        if not os.path.isfile(self._path):
            # file doesn't exist yet
            return

        with open(self._path, 'r', encoding='utf-8') as fp:
            j = json.load(fp)
            self._last_tester_initials = j['last_tester_initials']
            self._last_position_x = j['last_position_x']
            self._last_position_y = j['last_position_y']



import reportlab.lib.pagesizes
from reportlab.lib import colors
from reportlab.lib.enums import TA_CENTER
from reportlab.lib.enums import TA_LEFT
from reportlab.lib.enums import TA_RIGHT
from reportlab.lib.pagesizes import landscape
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.lib.units import cm
from reportlab.lib.units import inch
from reportlab.platypus import Frame
from reportlab.platypus import PageTemplate
from reportlab.platypus import Paragraph
from reportlab.platypus import SimpleDocTemplate
from reportlab.platypus import Table
from reportlab.platypus import TableStyle

from medver_pytest import services
from medver_pytest.constants import Constants
from medver_pytest.utils import Utils


# -------------------
## Base class for generating a PDF
class GenBasePdf:  # pylint: disable=too-few-public-methods

    # -------------------
    ## constructor
    def __init__(self):
        ## holds reference to the current PDF doc
        self._doc = None
        ## holds the list of style sheets for this doc
        self._style_sheet = None
        ## holds the current list of elements in the doc
        self._elements = []

        # uncomment to debug
        # services.logger.dbg(f"page_info: {services.cfg.page_info}")

    # -------------------
    ## initialize dcoument
    #
    # @param path   the path to the document
    # @return None
    def _doc_init(self, path):
        if services.cfg.page_info.page_size == 'letter':
            psize = reportlab.lib.pagesizes.LETTER
        elif services.cfg.page_info.page_size == 'a4':
            psize = reportlab.lib.pagesizes.A4
        else:
            services.abort(f'unknown page_size: {services.cfg.page_info.page_size}')
            return

        if services.cfg.page_info.orientation == 'portrait':
            self._doc = SimpleDocTemplate(path,
                                          pagesize=psize,
                                          leftMargin=0.5 * inch,
                                          rightMargin=0.5 * inch,
                                          topMargin=0.75 * inch,
                                          bottomMargin=1.0 * inch,
                                          )
        else:
            # landscape
            self._doc = SimpleDocTemplate(path,
                                          pagesize=landscape(psize),
                                          leftMargin=0.5 * inch,
                                          rightMargin=0.5 * inch,
                                          topMargin=0.75 * inch,
                                          bottomMargin=1.0 * inch,
                                          )

        self._style_sheet = getSampleStyleSheet()
        # set the overall font sizes
        # Normal is used for headers/footers
        self._style_sheet['Normal'].fontSize = 8
        # BodyText is used for all others
        self._style_sheet['BodyText'].fontSize = 8
        self._style_sheet.add(ParagraphStyle(name='ExtraSpace',
                                             parent=self._style_sheet['Normal'],
                                             alignment=TA_LEFT,
                                             spaceAfter=5))
        self._style_sheet.add(ParagraphStyle(name='LeftPara',
                                             parent=self._style_sheet['Normal'],
                                             alignment=TA_LEFT))
        self._style_sheet.add(ParagraphStyle(name='CenterPara',
                                             parent=self._style_sheet['Normal'],
                                             alignment=TA_CENTER))
        self._style_sheet.add(ParagraphStyle(name='RightPara',
                                             parent=self._style_sheet['Normal'],
                                             alignment=TA_RIGHT))

        if services.cfg.page_info.orientation == 'portrait':
            self._style_sheet.add(ParagraphStyle(name='Mono',
                                                 parent=self._style_sheet['BodyText'],
                                                 alignment=TA_LEFT,
                                                 fontName='Courier-Bold',
                                                 fontSize=6
                                                 ))
        else:
            self._style_sheet.add(ParagraphStyle(name='Mono',
                                                 parent=self._style_sheet['BodyText'],
                                                 alignment=TA_LEFT,
                                                 fontName='Courier-Bold',
                                                 fontSize=6
                                                 ))

        self._elements = []
        frameh = Frame(self._doc.leftMargin,  # pylint: disable=no-member
                       self._doc.bottomMargin,  # pylint: disable=no-member
                       self._doc.width,
                       self._doc.height - 2 * cm,
                       id='normal')
        pt = PageTemplate(id='test', frames=[frameh], onPage=self._header_and_footer)
        self._doc.addPageTemplates([pt])

    # -------------------
    ## callback for generating header and footer on the given canvas
    #
    # @param canvas  the canvas to draw the header/footer into
    # @param doc     the doc
    # @return None
    def _header_and_footer(self, canvas, doc):
        self._header(canvas, doc)
        self._footer(canvas, doc)

    # -------------------
    ## callback for generating header on the given canvas
    #
    # @param canvas  the canvas to draw the header into
    # @param doc     the doc
    # @return None
    def _header(self, canvas, doc):
        canvas.saveState()
        edge_width = 1.75 * inch
        if services.cfg.page_info.orientation == 'portrait':
            t = Table([self._get_data('header', canvas)],
                      # should add to 7.5"
                      colWidths=[
                          edge_width,  # left
                          4.00 * inch,  # middle
                          edge_width,  # right
                      ],  # noqa
                      spaceBefore=0.1,
                      spaceAfter=0.1,
                      rowHeights=0.5 * inch,
                      )
        else:
            t = Table([self._get_data('header', canvas)],
                      # should add to 10"
                      colWidths=[
                          edge_width,  # left
                          6.50 * inch,  # middle
                          edge_width,  # right
                      ],  # noqa
                      spaceBefore=0.1,
                      spaceAfter=0.1,
                      rowHeights=0.5 * inch,
                      )

        t.setStyle(TableStyle(
            [
                # border below only
                ('LINEBELOW', (0, 0), (-1, -1), 0.25, colors.black),
                # uncomment to get field delimiters in headers/footers
                # ('INNERGRID', (0, 0), (-1, -1), 0.25, colors.black),

                ('VALIGN', (0, 0), (-1, -1), 'MIDDLE'),
            ]
        ))

        _, height = t.wrap(doc.width, doc.topMargin)
        t.drawOn(canvas, doc.leftMargin, doc.height + doc.bottomMargin + doc.topMargin - height)
        canvas.restoreState()

    # -------------------
    ## callback for generating footer on the given canvas
    #
    # @param canvas  the canvas to draw the footer into
    # @param doc     the doc
    # @return None
    def _footer(self, canvas, doc):
        canvas.saveState()

        edge_width = 1.75 * inch
        if services.cfg.page_info.orientation == 'portrait':
            t = Table([self._get_data('footer', canvas)],
                      # should add to 7.5"
                      colWidths=[
                          edge_width,  # left
                          4.00 * inch,  # middle
                          edge_width,  # right
                      ],  # noqa
                      spaceBefore=0.0,
                      spaceAfter=0.0,
                      )
        else:
            t = Table([self._get_data('footer', canvas)],
                      # should add to 10"
                      colWidths=[
                          edge_width,  # left
                          6.50 * inch,  # middle
                          edge_width,  # right
                      ],  # noqa
                      spaceBefore=0.1,
                      spaceAfter=0.1,
                      rowHeights=0.5 * inch,
                      )

        t.setStyle(TableStyle(
            [
                # border below only
                ('LINEABOVE', (0, 0), (-1, -1), 0.25, colors.black),
                # uncomment to get field delimiters in headers/footers
                # ('INNERGRID', (0, 0), (-1, -1), 0.25, colors.black),

                ('VALIGN', (0, 0), (-1, -1), 'MIDDLE'),
            ]
        ))

        _, height = t.wrap(doc.width, doc.bottomMargin)
        t.drawOn(canvas, doc.leftMargin, height)

        canvas.restoreState()

    # -------------------
    ## callback for generating header/footer on the given canvas
    #
    # @param tag     the key into the cfg JSON object
    # @param canvas  the canvas to draw the footer into
    # @return the current row info for the header/footer
    def _get_data(self, tag, canvas):
        row = []
        if services.cfg.page_info[tag].left == '<pageno>':
            row.append(Paragraph(f'Page {canvas.getPageNumber()}', self._style_sheet['LeftPara']))
        else:
            row.append(Paragraph(services.cfg.page_info[tag].left, self._style_sheet['LeftPara']))

        if services.cfg.page_info[tag].middle == '<pageno>':
            row.append(Paragraph(f'Page {canvas.getPageNumber()}', self._style_sheet['CenterPara']))
        else:
            row.append(Paragraph(services.cfg.page_info[tag].middle, self._style_sheet['CenterPara']))

        if services.cfg.page_info[tag].right == '<pageno>':
            row.append(Paragraph(f'Page {canvas.getPageNumber()}', self._style_sheet['LeftPara']))
        else:
            row.append(Paragraph(services.cfg.page_info[tag].right, self._style_sheet['LeftPara']))

        return row

    # -------------------
    ## build the document with the current list of elements
    #
    # @return None
    def _build(self):
        self._doc.build(self._elements,
                        onFirstPage=self._header_and_footer,
                        onLaterPages=self._header_and_footer)

    # -------------------
    ## generate test run information
    #
    # @return None
    def _gen_test_run_details(self):
        para = Paragraph('<b>Test Run Details</b>', self._style_sheet['BodyText'])
        self._elements.append(para)

        line = f"{'Test Run Type': <20}: {services.cfg.test_run_type}"
        para = Paragraph(line, bulletText='\u2022')
        self._elements.append(para)

        line = f"{'Test Run ID': <20}: {services.cfg.test_run_id}"
        para = Paragraph(line, bulletText='\u2022')
        self._elements.append(para)

        dts = Utils.get_dts(use_cfg_fmt=True)
        line = f"{'Document Generated': <20}: {dts}"
        para = Paragraph(line, bulletText='\u2022')
        self._elements.append(para)

        line = f"{'medver-pytest version'}: v{Constants.version}"
        para = Paragraph(line, bulletText='\u2022')
        self._elements.append(para)

    # -------------------
    ## generate title
    #
    # @param title  the title to draw
    # @return None
    def _gen_title(self, title):
        para = Paragraph(f'<b>{title}</b>', self._style_sheet['BodyText'])
        self._elements.append(para)



import json
import os
import re

import jsmin

from . import services
from .build_info import BuildInfo
from .cli import Cli
from .constants import Constants
from medver_pytest.lib.report.common.page_info import PageInfo


# -------------------
## Holds configuration functions and globals
class Cfg:
    # -------------------
    ## constructor
    def __init__(self):
        # --- Public - default settings
        # Note: order these for order in report()

        # == location of dirs and files
        ## holds path to cfg file
        self.cfg_path = None
        ## holds path to output directory
        self.outdir = 'out'
        ## holds path to directory holding status.json file
        self.statusdir = '.'
        ## holds path to the requirements json file
        self.reqmt_json_path = None

        # == test run related
        ## holds run type: one of formal, dryrun, dev
        self.test_run_type = 'dev'
        ## holds run id
        self.test_run_id = 'dev-001'

        ## report types to generate, valid: txt, pdf, docx, min-html
        self.report_types = ['txt', 'pdf', 'docx', 'min-html']
        ## report mode; use to suppress the creation of JSON files or not
        self.report_mode = False

        ## holds storage type: one of local, shared
        self.storage_type = 'local'
        ## holds path to the shared directory to publish; used only if storage_type is shared
        self.storage_shared_dir = None

        ## holds format to use for DTS
        self.dts_format = "%Y-%m-%d %H:%M:%S %Z"
        ## flag indicates to use UTC or local time
        self.use_utc = False

        # === skip these in report()

        ## holds page info e.g. headers and footers
        self.page_info = PageInfo()
        ## holds current test name
        self.test_script = None
        ## list of manual testers
        self.testers = []
        ## internal IUV use only; indicates if in IUV mode
        self.iuvmode = False

        ## holds reference to the CLI object
        self._cli = Cli()

    # -------------------
    ## add options for conftest.py behavior
    #
    # @param parser the pytest parser object
    # @return None
    def cli_addoption(self, parser):
        self._cli.addoption(parser)

    # -------------------
    ## handle the incoming CLI flags for conftest.py behavior
    #
    # @param config the pytest config reference
    # @return None
    def cli_configure(self, config):
        self._cli.configure(config)

    # -------------------
    ## add option for CLI behavior
    #
    # @return None
    def cli_parse(self):
        self._cli.parse()

    # -------------------
    ## used to set command line interface switches
    #
    # @param name   the name of the variable
    # @param value  the value of the variable to use
    # @return None
    def cli_set(self, name, value):
        if self._cli.is_valid(name):
            setattr(self, name, value)

    # -------------------
    ## initialize - step1
    # read cfg json file
    #
    # @param report_mode used to suppress creation of out/*.json files (for reporting)
    # @return None
    def init(self, report_mode):
        # save cfg_path and iuvmode if given on CLI
        self._cli.save()

        if self.iuvmode:
            services.harness.init_iuv()
            if self.cfg_path is None:
                # coverage: iuvmode is only set during IUV and UT runs
                # default to iuv/cfg0.json
                self.cfg_path = os.path.join('lib/iuv', 'cfg0.json')

        # if still not set, default to cfg.json
        if self.cfg_path is None:
            self.cfg_path = 'cfg.json'

        self._read_ini()
        self.report_mode = report_mode

        # override any values given on CLI
        self._cli.override()

        if not os.path.isdir(self.outdir):  # pragma: no cover
            # coverage: in IUV and UT, outdir is created by scripts
            os.mkdir(self.outdir)

        self._check()

    # -------------------
    ## initialize - step2
    # get the current test name
    #
    # @return None
    def init2(self):
        self.test_script = None
        if 'PYTEST_CURRENT_TEST' in os.environ:  # pragma: no cover
            # coverage: in IUV and UT, variable is always set
            # e.g.  ver/test_sample_ver1.py
            m = re.search(r'test_(\w+)\.py::(\w+)', os.getenv('PYTEST_CURRENT_TEST'))
            if m:
                self.test_script = m.group(1)

    # -------------------
    ## report configuration to the log
    #
    # @return None
    def report(self):
        services.logger.start('Cfg:')
        for name, val in vars(self).items():
            self._log_attr(name, val)

        services.logger.line(f"  {'medver-pytest Version': <20}: {Constants.version}")
        services.logger.line(f"  {'git SHA': <20}: {BuildInfo.git_sha}")
        services.logger.line(f"  {'git Branch': <20}: {BuildInfo.git_branch}")
        services.logger.line(f"  {'git Uncommitted': <20}: {BuildInfo.git_uncommitted}")
        services.logger.line(f"  {'git Unpushed': <20}: {BuildInfo.git_unpushed}")

        services.logger.line('')

    # -------------------
    ## log configuration attribute
    #
    # @param name   the attribute name
    # @param val    the attribute value
    # @return None
    def _log_attr(self, name, val):
        if name.startswith('_'):
            # skip private attributes
            return

        if name == 'iuvmode' and not self.iuvmode:
            # only report if in iuvmode
            return

        if name in ['test_script', 'testers', 'page_info']:
            return

        if name == 'outdir':
            name = 'output_dir'
        elif name == 'statusdir':
            name = 'status_dir'

        elems = []
        for elem in name.split('_'):
            if elem in ['id', 'utc', 'json', 'dts']:
                elems.append(elem.upper())
            else:
                elems.append(elem.capitalize())

        nice_name = ' '.join(elems)
        services.logger.line(f'  {nice_name: <20}: {val}')

    # -------------------
    ## read the cfg json file
    # set attributes in this class based on content
    #
    # @return None
    def _read_ini(self):
        if not os.path.isfile(self.cfg_path):
            services.logger.warn(f'cfg_path "{self.cfg_path}" not found')
            return

        # load json file
        with open(self.cfg_path, 'r', encoding='utf-8') as fp:
            cleanj = jsmin.jsmin(fp.read())
            j = json.loads(cleanj)

        # override and/or add to available attributes
        for key, value in j.items():
            if key == 'tp_report':
                self.page_info.init_tp_report(value)
            elif key == 'tp_protocol':
                self.page_info.init_tp_protocol(value)
            elif key == 'trace':
                self.page_info.init_trace(value)
            elif key == 'summary':
                self.page_info.init_summary(value)
            else:
                setattr(self, key, value)

    # -------------------
    ## checks all values for valid content
    # calls abort() if there are any failures
    #
    # @return None
    def _check(self):
        errs = 0
        errs += self.page_info.check()

        if self.cfg_path is None:
            errs += 1
            services.logger.err('cfg_path is None; must be a valid path to a cfg json file')

        if not os.path.isdir(self.outdir):
            errs += 1
            services.logger.err(f'outdir is not a valid dir; does it exist?: {self.outdir}')

        valid = ['local', 'shared']
        if self.storage_type not in valid:
            errs += 1
            services.logger.err(f'storage_type can only be {valid}, found: {self.storage_type}')
        elif self.storage_type == 'shared':
            if not os.path.isdir(self.storage_shared_dir):
                errs += 1
                services.logger.err(f'storage_shared_dir is not a valid dir; does it exist?: {self.storage_shared_dir}')

        # report_mode: nothing to check

        valid = ['formal', 'dryrun', 'dev']
        if self.test_run_type not in valid:
            errs += 1
            services.logger.err(f'test_run_type can only be {valid}, found: {self.test_run_type}')

        if self.test_run_id is None:
            errs += 1
            services.logger.err('test_run_id is None; must be a valid test run id')

        if self.dts_format is None:
            errs += 1
            services.logger.err('dts_format is None; must be a valid DTS format')
            # TODO check for other invalid formats. How?

        if self.use_utc is None:
            errs += 1
            services.logger.err('use_utc is None; must be True or False')

        # page_info: nothing to check
        # test_script: nothing to check
        # reqmt_json_path: nothing to check
        # tp_protocol_fname: nothing to check
        # tp_report_fname: nothing to check

        valid = ['txt', 'pdf', 'docx', 'min-html']
        if self.report_types is None:
            errs += 1
            services.logger.err('report_types is None; must be a list')
        elif not self.report_types:
            # allow an empty list
            pass
        else:
            # non-empty list, check all entries
            for rtype in self.report_types:
                if rtype not in valid:
                    errs += 1
                    services.logger.err(f'report_types can only be {valid}, found: {rtype}')

        if errs > 0:
            services.abort('cfg error found')



from . import services
from .cfg import Cfg
from .constants import Constants
from medver_pytest.lib.iuv.iuv import IUV
from medver_pytest.lib.log.logger import Logger
from medver_pytest.lib.log import LoggerStdout
from .protocol import Protocol
from medver_pytest.lib.report import Report
from .status_info import StatusInfo
from medver_pytest.lib.storage import Storage
from .summary import Summary
from .trace_matrix import TraceMatrix
from .verifier import Verifier


# -------------------
## Holds the overall test harness
# initializes the services object
class PytestHarness:
    # -------------------
    ## constructor
    def __init__(self):
        services.harness = self

        ## holds the IUV object when needed
        self.iuv = None
        ## holds reference to the protocol object
        self.proto = None
        ## holds reference to the verifier object
        self.ver = None

        services.logger = LoggerStdout()
        services.logger.init()

        services.cfg = Cfg()

    # -------------------
    ## initialize IUV components
    #
    # @return None
    def init_iuv(self):
        if services.cfg.iuvmode:  # pragma: no cover
            self.iuv = IUV()
            self.iuv.init()

    # -------------------
    ## initialize - once per invocation
    #
    # @param report_mode used to suppress creation of out/*.json files (for reporting)
    # @return None
    def init(self, report_mode=False):
        services.cfg.init(report_mode)

        # after cfg indicates where log files are stored, can use normal logger
        services.logger = Logger()
        if report_mode:
            # use a different filename for report logging
            services.logger.init(fname='pytest_report.log')
        else:
            services.logger.init()

        services.storage = Storage.factory()
        services.summary = Summary()
        services.trace = TraceMatrix()
        services.proto = Protocol()
        services.status = StatusInfo()

        self.proto = services.proto
        self.ver = Verifier()

        services.cfg.init2()
        services.cfg.report()

        services.proto.init()
        services.storage.init()

        if services.cfg.iuvmode:  # pragma: no cover
            # coverage: iuvmode is only set during IUV and UT runs
            self.iuv.init2()

    # -------------------
    ## gives access to the cfg from reports etc.
    #
    # @return the services.cfg object
    @property
    def cfg(self):
        return services.cfg

    # -------------------
    ## gives access to the logger from reports etc.
    #
    # @return the services.logger object
    @property
    def logger(self):
        return services.logger

    # -------------------
    ## returns the current version
    #
    # @return the version string
    @property
    def version(self):
        return Constants.version

    # -------------------
    ## terminate
    #
    # @return None
    def term(self):
        services.proto.term()
        services.trace.term()
        services.summary.term()
        services.storage.term()

    # -------------------
    ## run a report
    #
    # @return None
    def report(self):
        rep = Report()
        rep.report()

    # -------------------
    ## abort the run
    #
    # @return None
    def abort(self):
        services.abort()



from medver_pytest import services


# -------------------
## Holds data for the Summary report
class Summary:
    # -------------------
    ## constructor
    def __init__(self):
        ## holds the summary info
        self._summary = None
        ## holds the current test script name
        self._current_test_script = None
        ## indicates if there is any information to be saved to the output file
        self._dirty = False

        self._summary = {
            'reqids': {},
            'protoids': {},
        }

    # -------------------
    ## terminate - save all content
    #
    # @return None
    def term(self):
        self.save()

    # -------------------
    ## clear current summary content
    #
    # @return None
    def _clear(self):
        self._summary = {
            'reqids': {},
            'protoids': {},
        }
        self._dirty = True

    # -------------------
    ## ensure that the current data is cleared out when the test script changes
    #
    # @return None
    def check_test_script(self):
        # if the test script name has changed, save the current content and re-init protocols info
        if self._current_test_script != services.cfg.test_script:
            if self._current_test_script is not None:
                self.save()

            self._current_test_script = services.cfg.test_script
            self._clear()
            self._dirty = False

    # -------------------
    ## add a result info to the summary
    # save to storage
    #
    # @param reqids    list of requirement ids for this result
    # @param protoid   the protocol id for this result
    # @param result    the result info
    # @return None
    def add_result(self, reqids, protoid, result):
        self.check_test_script()

        if reqids is not None:
            for reqid in reqids:
                if reqid in self._summary['reqids']:
                    self._reqid_add_to(reqid, result)
                else:
                    self._reqid_add(reqid, result)

        if protoid in self._summary['protoids']:
            self._protoid_add_to(protoid, result)
        else:
            self._protoid_add(protoid, result)

        self.save()

    # -------------------
    ## add a reqid and result to the summary info
    #
    # @param reqid    rthe requirement id
    # @param result   the result i.e. PASS/FAIL
    # @return None
    def _reqid_add(self, reqid, result):
        self._summary['reqids'][reqid] = {
            'count': 1,
            'result': result
        }
        self._dirty = True

    # -------------------
    ## add result to an existing reqid info
    # if the current list is all PASS and the current result is a FAIL,
    # then the overall result is FAIL, otherwise the result remains as it was
    #
    # @param reqid    the requirement id
    # @param result   the result i.e. PASS/FAIL
    # @return None
    def _reqid_add_to(self, reqid, result):
        self._summary['reqids'][reqid]['count'] += 1
        if self._summary['reqids'][reqid]['result'] == 'PASS':
            self._summary['reqids'][reqid]['result'] = result
        # else if FAIL, stays FAIL

        self._dirty = True

    # -------------------
    ## add a protoid and result to the summary info
    #
    # @param protoid  the protocol id
    # @param result   the result i.e. PASS/FAIL
    # @return None
    def _protoid_add(self, protoid, result):
        self._summary['protoids'][protoid] = {
            'count': 1,
            'result': result
        }
        self._dirty = True

    # -------------------
    ## add result to an existing protoid info
    # if the current list is all PASS and the current result is a FAIL,
    # then the overall result is FAIL, otherwise the result remains as it was
    #
    # @param protoid   the protocol id
    # @param result    the result i.e. PASS/FAIL
    # @return None
    def _protoid_add_to(self, protoid, result):
        self._summary['protoids'][protoid]['count'] += 1
        if self._summary['protoids'][protoid]['result'] == 'PASS':
            self._summary['protoids'][protoid]['result'] = result
        # else if FAIL, stays FAIL

        self._dirty = True

    # -------------------
    ## save the summary info to storage
    #
    # @return None
    def save(self):
        services.storage.save_summary(self._summary)
        self._dirty = False

    # ===  IUV only

    # -------------------
    ## (IUV only) get current dirty flag
    #
    # @return dirty flag
    def iuv_get_dirty(self):
        return self._dirty

    # -------------------
    ## (IUV only) get summary object
    #
    # @return summary object
    def iuv_get_summary(self):
        return self._summary



import os

from .gen_summary_base import GenSummaryBase
from ..common.gen_base_txt import GenBaseTxt
from medver_pytest import services
from medver_pytest.constants import Constants


# -------------------
## Generates a Summary report in text format
class GenSummaryTxt(GenSummaryBase, GenBaseTxt):
    # -------------------
    ## constructor
    #
    # @param summary    the summary data to use
    def __init__(self, summary):
        GenBaseTxt.__init__(self)
        GenSummaryBase.__init__(self, summary)

        ## holds path to the output file
        self._path = os.path.join(services.cfg.outdir, f'{Constants.summary_fname}.txt')

        ## holds file pointer to the output file
        self._fp = None

    # -------------------
    ## generate the report
    #
    # @return None
    def gen(self):
        self._gen_init()

        with open(self._path, 'w', encoding='UTF-8') as self._fp:
            self._gen_test_run_details()
            self._gen_title('Summary')

            self._gen_req_summary()
            self._fp.write('\n')
            self._gen_proto_summary()

    # -------------------
    ## generate requirements summary
    #
    # @return None
    def _gen_req_summary(self):
        # total requirements
        if self._requirements.is_found:
            total = self._requirements.total
            self._write_line('#Requirements', total, total)

            count = self._requirements.total_manual
            self._write_line('Manual Requirements', count, total)

            count = self._requirements.total_both
            self._write_line('Manual & Auto Requirements', count, total)

            total = self._requirements.total_automated
        else:
            total = len(self._summary['reqids'])

        self._write_line('Automated Requirements', total, total)

        # count failing/passing reqmts
        num_failing, num_passing, num_invalid, num_missing = self._count_requirements(total, self._report_invalid)
        self._write_line('   Invalid', num_invalid, total)
        self._write_line('   PASS', num_passing, total)
        self._write_line('   FAIL', num_failing, total)
        self._write_line('   Not tested', num_missing, total)

    # -------------------
    ## callback to report any invalid requirements
    #
    # @param reqid  the reqid to report
    # @return None
    def _report_invalid(self, reqid):
        msg = f'   {reqid} not found in {services.cfg.reqmt_json_path}'
        services.logger.warn(msg)
        self._fp.write(f'{msg}\n')

    # -------------------
    ## generate protocol summary
    #
    # @return None
    def _gen_proto_summary(self):
        # total protocols
        total = len(self._summary['protoids'])
        self._write_line('#Protocols', total, total)

        # count failing/passing reqmts
        num_failing, num_passing = self._count_protocols()
        self._write_line('   PASS', num_passing, total)
        self._write_line('   FAIL', num_failing, total)

    # -------------------
    ## write a line to the given file
    #
    # @param msg    the line header to write
    # @param count  the value to write
    # @param total  the total count
    # @return None
    def _write_line(self, msg, count, total):
        pct = self._pct(count, total)
        self._fp.write(f'{msg: <27}: {count: >3}  {pct}\n')



from medver_pytest import services
from medver_pytest.constants import Constants
from medver_pytest.utils import Utils


# -------------------
## Base class for generating a text file
class GenBaseTxt:  # pylint: disable=too-few-public-methods

    # -------------------
    ## constructor
    def __init__(self):
        ## holds file pointer for the generated text file
        self._fp = None

    # -------------------
    ## generate test run information
    #
    # @return None
    def _gen_test_run_details(self):
        self._gen_title2('Test Run Details')
        self._fp.write(f"{'Test Run Type': <20}: {services.cfg.test_run_type}\n")
        self._fp.write(f"{'Test Run ID': <20}: {services.cfg.test_run_id}\n")
        dts = Utils.get_dts(use_cfg_fmt=True)
        self._fp.write(f"{'Document Generated': <20}: {dts}\n")
        self._fp.write(f"{'medver-pytest version': <20}: v{Constants.version}\n")

        services.cfg.page_info.report(self._fp)
        self._fp.write('\n')

    # -------------------
    ## generate title
    #
    # @param title  the title to draw
    # @return None
    def _gen_title(self, title):
        self._gen_title2(title)
        self._fp.write('\n')

    # -------------------
    ## generate title lines
    #
    # @param title  the title to draw
    # @return None
    def _gen_title2(self, title):
        self._fp.write(f'{title}\n')
        self._fp.write(f"{'-' * len(title)}\n")



from medver_pytest import services
from medver_pytest.requirements import Requirements


# -------------------
## Common functions for Trace Matrix reports
class GenTraceMatrixBase:  # pylint: disable=too-few-public-methods
    # -------------------
    ## constructor
    #
    # @param matrix   the matrix info to use
    def __init__(self, matrix):
        ## holds reference to the matrix info
        self._matrix = matrix
        ## holds reference to the Requirents object
        self._requirements = None

    # -------------------
    ## initialization
    #
    # @return None
    def _gen_init(self):
        services.logger.start('report: trace matrix')

        self._requirements = Requirements()
        self._requirements.load()

    # -------------------
    ## add any extra reqmts in the test protocol data
    # to the requirement list
    #
    # @return None
    def _load_matrix_requirements(self):
        for req_id in self._matrix.keys():
            self._requirements.add_reqid(req_id)

    # -------------------
    ## iterates through all requirements and creates a row for each entry as needed
    #
    # @param report_header   callback to report a header line for this requirement
    # @param report_desc     callback to report the current requirement description
    # @param report_details  callback to report the details for this requirement
    # @return None
    def _report_requirements(self, report_header, report_desc, report_details):
        for req_id, reqmt in self._requirements.all_reqids():
            report_header(req_id, f'({reqmt["tester"]})')
            report_desc(reqmt['desc'])
            count = 0
            if req_id in self._matrix:
                for proto in sorted(self._matrix[req_id], key=lambda item: item['proto_id']):
                    report_details(proto['proto_id'], proto['proto_info'])
                    count += 1
                    if count > 35:
                        count = 0
                        # finish off current row
                        report_details(None, None)

                        # start a new one
                        report_header(req_id, f'({reqmt["tester"]})\n(cont\'d)')
                        report_desc(reqmt['desc'])

            if count == 0:
                if reqmt["tester"] == 'manual':
                    report_details(None, 'N/A')
                else:
                    report_details(None, '<b>ERR missing test case for this requirement</b>')

            # blank line or go to next row of the table
            report_details(None, None)



from .constants_version import ConstantsVersion


# -------------------
## Holds various constants
class Constants(ConstantsVersion):  # pylint: disable=too-few-public-methods
    ## trace document file name
    trace_fname = 'trace'

    ## summary document file name
    summary_fname = 'summary'

    ## test report document file name - with results
    tp_report_fname = 'test_report'

    ## test protocol document file name - no results
    tp_protocol_fname = 'test_protocol'



import os

from reportlab.lib import colors
from reportlab.lib.units import inch
from reportlab.platypus import Paragraph
from reportlab.platypus import Table
from reportlab.platypus import TableStyle

from .gen_summary_base import GenSummaryBase
from ..common.gen_base_pdf import GenBasePdf
from medver_pytest import services
from medver_pytest.constants import Constants


# -------------------
## Generates a Summary report in PDF format
class GenSummaryPdf(GenBasePdf, GenSummaryBase):
    # -------------------
    ## constructor
    #
    # @param summary    the summary data to use
    def __init__(self, summary):
        GenBasePdf.__init__(self)
        GenSummaryBase.__init__(self, summary)

        ## holds reference to the table being created
        self._tbl = None

        path = os.path.join(services.cfg.outdir, f'{Constants.summary_fname}.pdf')
        self._doc_init(path)

    # -------------------
    ## generate the report
    #
    # @return None
    def gen(self):
        self._gen_init()

        self._gen_test_run_details()
        self._gen_title('Summary')

        self._gen_req_summary()
        self._gen_proto_summary()

        self._build()

    # -------------------
    ## generate requirements summary
    #
    # @return None
    def _gen_req_summary(self):
        self._tbl = []
        self._tbl.append([
            Paragraph('<b>Requirements Summary</b>', self._style_sheet['BodyText'])
        ])

        # total requirements
        self._tbl.append(['', 'Count', 'Percentage'])

        # total requirements
        total = self._get_total_requirements()
        self._write_line('#Requirements', total, total)

        num_failing, num_passing, num_invalid, num_missing = self._count_requirements(total, self._report_invalid)
        self._write_line('   Invalid', num_invalid, total)
        self._write_line('   PASS', num_passing, total)
        self._write_line('   FAIL', num_failing, total)
        self._write_line('   Not tested', num_missing, total)

        self._gen_summary_table(self._tbl)

    # -------------------
    ## callback to report any invalid requirements
    #
    # @param reqid   the reqid to report as invalid
    # @return None
    def _report_invalid(self, reqid):
        msg = f'   {reqid} not found in {services.cfg.reqmt_json_path}'
        self._tbl.append([msg, '', ''])
        services.logger.warn(msg)

    # -------------------
    ## generate protocol summary
    #
    # @return None
    def _gen_proto_summary(self):
        self._tbl = []
        self._tbl.append([
            Paragraph('<b>Protocol Summary</b>', self._style_sheet['BodyText'])
        ])

        # total requirements
        self._tbl.append(['', 'Count', 'Percentage'])

        total = len(self._summary['protoids'])
        self._write_line('#Protocols', total, total)

        # count failing/passing reqmts
        num_failing, num_passing = self._count_protocols()
        self._write_line('   PASS', num_passing, total)
        self._write_line('   FAIL', num_failing, total)

        self._gen_summary_table(self._tbl)

    # -------------------
    ## write a line to table
    #
    # @param msg    the line header to write
    # @param count  the value to write
    # @param total  the total count
    # @return None
    def _write_line(self, msg, count, total):
        pct = self._pct(count, total)
        self._tbl.append([msg, count, pct])

    # -------------------
    ## generate a summary table
    #
    # @param tbl   the table to generate into
    # @return None
    def _gen_summary_table(self, tbl):
        tbl_widths = [
            3.20 * inch,  # desc
            0.50 * inch,  # value
            0.85 * inch,  # pct
        ]

        t = Table(tbl,
                  colWidths=tbl_widths,
                  spaceBefore=0.25 * inch,
                  spaceAfter=0.25 * inch,
                  )

        t.setStyle(TableStyle(
            [
                # borders
                ('BOX', (0, 0), (-1, -1), 0.25, colors.black),
                ('INNERGRID', (0, 0), (-1, -1), 0.25, colors.black),

                # top row spans all columns
                ('BACKGROUND', (0, 0), (2, 0), colors.lightblue),
                ('SPAN', (0, 0), (2, 0)),

                # 2nd row is a title row
                ('BACKGROUND', (0, 1), (2, 1), colors.lightgrey),

                # "description" column is middle/center
                ('VALIGN', (0, 0), (0, -1), 'MIDDLE'),
                ('ALIGN', (0, 0), (0, -1), 'LEFT'),

                # "value" column is middle/right
                ('VALIGN', (1, 0), (1, -1), 'MIDDLE'),
                ('ALIGN', (1, 0), (1, -1), 'RIGHT'),

                # "%" column is middle/right
                ('VALIGN', (2, 0), (2, -1), 'MIDDLE'),
                ('ALIGN', (2, 0), (2, -1), 'RIGHT'),
            ]
        ))
        self._elements.append(t)



import os

from docx import Document
from docx.shared import Inches

from .gen_trace_matrix_base import GenTraceMatrixBase
from ..common.gen_base_docx import GenBaseDocx
from medver_pytest import services
from medver_pytest.constants import Constants


# -------------------
## Generates a Trace Matrix report in docx format
class GenTraceMatrixDocx(GenBaseDocx, GenTraceMatrixBase):
    # -------------------
    ## constructor
    #
    # @param matrix   the data to use
    def __init__(self, matrix):
        GenBaseDocx.__init__(self)
        GenTraceMatrixBase.__init__(self, matrix)

        ## holds reference to the current table being generated
        self._tbl = None
        ## holds current reqmt id for the row
        self._row_id = None
        ## holds current reqmt info for the row
        self._row_info = None

    # -------------------
    ## generate the report
    #
    # @return None
    def gen(self):
        self._gen_init()

        ## see GenBaseDocx
        self._doc = Document()
        self._doc_init()

        self._gen_test_run_details()
        self._gen_title('Trace Matrix')

        self._gen_trace_table()

        self._build()

    # -------------------
    ## generate a trace matrix table
    #
    # @return None
    def _gen_trace_table(self):
        self._create_table()
        self._gen_table_header_row()

        self._load_matrix_requirements()
        ## see base class doc
        self._report_requirements(self._report_header, self._report_desc, self._report_details)

    # -------------------
    ## callback to report the current requirement and tester info
    #
    # @param req_id   the requirement id to report
    # @param info     the info to report
    # @return None
    def _report_header(self, req_id, info):
        if self._row_id is not None:
            row_cells = self._tbl.add_row().cells
            self._set_cell_id(row_cells[0])
            self._set_cell_info(row_cells[1])

        self._row_id = f'{req_id}\n{info}'
        self._row_info = []

    # -------------------
    ## callback to report the current requirement's description (if available)
    #
    # @param desc     the description to report
    # @return None
    def _report_desc(self, desc):
        self._row_info.append(desc)

    # -------------------
    ## callback to report protocol id and info for the current requirement
    #
    # @param proto_id     the protocol id to report
    # @param proto_info   the protocol info to report
    # @return None
    def _report_details(self, proto_id, proto_info):
        if proto_id is None and proto_info is None:
            row_cells = self._tbl.add_row().cells
            self._set_cell_id(row_cells[0])
            self._set_cell_info(row_cells[1])

            self._row_id = None
            self._row_info = None
        elif proto_id is None:
            self._row_info.append(proto_info)
        else:
            self._row_info.append(f'{proto_id} {proto_info}')

    # -------------------
    ## create a table with 2 columns
    #
    # @return None
    def _create_table(self):
        # TODO add 0.25 space before and after table
        #     t = Table(self._tbl,
        #               colWidths=tbl_widths,
        #               spaceBefore=0.25 * inch,
        #               spaceAfter=0.25 * inch,
        #               )
        self._tbl = self._doc.add_table(rows=0, cols=2)
        self._tbl.style = 'Table Grid'
        self._align_table_center(self._tbl)

        self._tbl.columns[0].width = Inches(0.75)  # reqid
        self._tbl.columns[1].width = Inches(6.25)  # proto_info

    # -------------------
    ## generate header2 row
    #
    # @return None
    def _gen_table_header_row(self):
        row = self._tbl.add_row().cells
        self._set_header_cell(row[0], 'Req.')
        self._set_header_cell(row[1], 'Protocol')

    # -------------------
    ## set header1 cell to the given text
    #
    # @param cell  the cell to set
    # @param text  the text to set the cell to
    # @return None
    def _set_header_cell(self, cell, text):
        cell.paragraphs[0].add_run(text)
        cell.paragraphs[0].style = 'ver_table1_header'
        self._set_cell_blue_background(cell)
        self._align_vertical_top(cell)

    # -------------------
    ## set the requirement id in the given cell
    #
    # @param cell  the cell to write to
    # @return None
    def _set_cell_id(self, cell):
        cell.text = ''
        cell.paragraphs[0].add_run(self._row_id)
        cell.paragraphs[0].style = 'ver_table1_cell'
        self._align_vertical_center(cell)
        cell.alignment = self._align_center()

    # -------------------
    ## set the protocol info in the given cell
    #
    # @param cell  the cell to write to
    # @return None
    def _set_cell_info(self, cell):
        cell.text = ''
        self._align_vertical_center(cell)
        cell.alignment = self._align_left()
        # self._row_info
        cell.paragraphs[0].style = 'ver_table1_cell'
        run = cell.paragraphs[0].add_run('Desc: ')
        run.italic = True
        run.bold = True
        run = cell.paragraphs[0].add_run(self._row_info[0])
        run.italic = True
        run.bold = False

        for i in range(1, len(self._row_info)):
            txt = self._row_info[i]
            if i == 1:
                txt = f'\n{txt}'

            boldness = False
            if txt.find('<b>') != -1 or txt.find('</b>') != -1 or txt.find('<b/>') != -1:
                boldness = True
                txt = txt.replace('<b>', '')
                txt = txt.replace('</b>', '')
                txt = txt.replace('<b/>', '')

            para = cell.add_paragraph(style='ver_table1_cell')
            run = para.add_run(txt)
            run.bold = boldness

    # -------------------
    ## build the document
    #
    # @return None
    def _build(self):
        path = os.path.join(services.cfg.outdir, f'{Constants.trace_fname}.docx')
        self._doc.save(path)



import os

from docx import Document
from docx.shared import Inches

from .gen_summary_base import GenSummaryBase
from ..common.gen_base_docx import GenBaseDocx
from medver_pytest import services
from medver_pytest.constants import Constants


# -------------------
## Generates a Summary report in PDF format
class GenSummaryDocx(GenBaseDocx, GenSummaryBase):
    # -------------------
    ## constructor
    #
    # @param summary    the summary data to use
    def __init__(self, summary):
        GenBaseDocx.__init__(self)
        GenSummaryBase.__init__(self, summary)

        ## holds reference to the table being created
        self._tbl = None

    # -------------------
    ## generate the report
    #
    # @return None
    def gen(self):
        self._gen_init()

        ## see GenBaseDocx
        self._doc = Document()
        self._doc_init()

        self._gen_test_run_details()
        self._gen_title('Summary')

        self._gen_req_summary()
        self._doc.add_paragraph('')
        self._gen_proto_summary()

        self._build()

    # -------------------
    ## generate requirements summary
    #
    # @return None
    def _gen_req_summary(self):
        self._create_table()

        # gen header rows
        self._gen_header1('Requirements Summary')
        self._gen_header2()

        # total requirements
        total = self._get_total_requirements()
        self._write_row('#Requirements', total, total)

        num_failing, num_passing, num_invalid, num_missing = self._count_requirements(total, self._report_invalid)
        self._write_row('   Invalid', num_invalid, total)
        self._write_row('   PASS', num_passing, total)
        self._write_row('   FAIL', num_failing, total)
        self._write_row('   Not tested', num_missing, total)

    # -------------------
    ## callback to report any invalid requirements
    #
    # @param reqid   the reqid to report as invalid
    # @return None
    def _report_invalid(self, reqid):
        msg = f'   {reqid} not found in {services.cfg.reqmt_json_path}'
        self._write_row(msg, None, None)
        services.logger.warn(msg)

    # -------------------
    ## generate protocol summary
    #
    # @return None
    def _gen_proto_summary(self):
        self._create_table()

        # gen header rows
        self._gen_header1('Protocol Summary')
        self._gen_header2()

        total = len(self._summary['protoids'])
        self._write_row('#Protocols', total, total)

        # count failing/passing reqmts
        num_failing, num_passing = self._count_protocols()
        self._write_row('   PASS', num_passing, total)
        self._write_row('   FAIL', num_failing, total)

    # -------------------
    ## create a table with 3 columns
    #
    # @return None
    def _create_table(self):
        # TODO add 0.25 space before and after table
        self._tbl = self._doc.add_table(rows=1, cols=3)
        self._tbl.style = 'Table Grid'
        self._align_table_center(self._tbl)
        self._tbl.columns[0].width = Inches(3.20)
        self._tbl.columns[1].width = Inches(0.65)
        self._tbl.columns[2].width = Inches(0.90)

    # -------------------
    ## generate header1 row
    #
    # @param title  the text to set the header to
    # @return None
    def _gen_header1(self, title):
        row = self._tbl.rows[0].cells
        self._set_header1_cell(row[0], title)
        row[0].merge(row[1])
        row[0].merge(row[2])

    # -------------------
    ## generate header2 row
    #
    # @return None
    def _gen_header2(self):
        row = self._tbl.add_row().cells
        self._set_header2_cell(row[0], '')
        self._set_header2_cell(row[1], 'Count')
        self._set_header2_cell(row[2], 'Percentage')

    # -------------------
    ## write a row to table
    #
    # @param msg    the line header to write
    # @param count  the value to write
    # @param total  the total count
    # @return None
    def _write_row(self, msg, count, total):
        if count is None or total is None:
            pct = ''
        else:
            pct = self._pct(count, total)
        if count is None:
            count = ''

        row = self._tbl.add_row().cells
        row[0].paragraphs[0].add_run(str(msg), )
        row[0].paragraphs[0].style = 'ver_table1_cell'

        row[1].paragraphs[0].add_run(f'{count: >8}', )
        row[1].paragraphs[0].style = 'ver_table1_cell'

        row[2].paragraphs[0].add_run(f'{pct: >8}', )
        row[2].paragraphs[0].style = 'ver_table1_cell'

    # -------------------
    ## set header1 cell to the given text
    #
    # @param cell  the cell to set
    # @param text  the text to set the cell to
    # @return None
    def _set_header1_cell(self, cell, text):
        cell.paragraphs[0].add_run(text)
        cell.paragraphs[0].style = 'ver_table1_header'
        self._set_cell_blue_background(cell)
        self._align_vertical_top(cell)

    # -------------------
    ## set header2 cell to the given text
    #
    # @param cell  the cell to set
    # @param text  the text to set the cell to
    # @return None
    def _set_header2_cell(self, cell, text):
        cell.paragraphs[0].add_run(text)
        cell.paragraphs[0].style = 'ver_table1_header'
        self._set_cell_grey_background(cell)
        self._align_vertical_top(cell)

    # -------------------
    ## build the document
    #
    # @return None
    def _build(self):
        path = os.path.join(services.cfg.outdir, f'{Constants.summary_fname}.docx')
        self._doc.save(path)



import docx
import docx.enum.table
import docx.enum.text
import docx.shared
from docx import Document

from .gen_tp_base import GenTpBase
from ..common.gen_base_docx import GenBaseDocx
from medver_pytest import services
from medver_pytest.result_summary import ResultSummary


# -------------------
## Generates a Test Protocol report in MSWord docx format
class GenTpDocx(GenBaseDocx, GenTpBase):
    # -------------------
    ## constructor
    #
    # @param protocols   the data to use
    # @param do_results  generate results or not
    def __init__(self, protocols, do_results=True):
        ## see base class doc
        self._testers = None
        ## see base clas doc
        self._doc_path = None

        GenBaseDocx.__init__(self)
        GenTpBase.__init__(self, protocols, do_results)

        self._init('docx', protocols, do_results)

    # -------------------
    ## generate the report
    #
    # @return None
    def gen(self):
        self._gen_init()
        # uncomment to debug:
        # services.logger.dbg(f'protocols: {json.dumps(self._protocols, indent=2)}')

        ## see GenBaseDocx
        self._doc = Document()
        self._doc_init()

        self._gen_test_run_details()
        title = self._get_title()
        self._gen_title(title)

        self._doc.add_paragraph('')
        for _, protocol in self._protocols.items():
            self._gen_protocol(protocol)

        self._build()

    # -------------------
    ## generate a protocol table using the given data
    #
    # @param protocol   the data to use
    # @return None
    def _gen_protocol(self, protocol):
        tbl = self._gen_protocol_table()
        self._gen_protocol_info(tbl, protocol)

        requirements = []
        need_space = self._gen_steps_info(tbl, protocol['steps'], requirements)
        self._update_protocol_table(tbl, need_space)

        self._tbl_add_space(1)
        self._gen_reqmt_table(requirements)

        if self._testers:
            self._tbl_add_space(1)
            self._gen_testers()

        self._tbl_add_space(2)

    # -------------------
    ## generate protocol info
    #
    # @param tbl        the table to write to
    # @param protocol   the data to use
    # @return None
    def _gen_protocol_info(self, tbl, protocol):
        self._gen_protocol_info_row0(tbl, protocol)
        self._gen_protocol_info_row1(tbl, protocol)
        self._gen_protocol_info_row2(tbl, protocol)
        self._gen_protocol_info_row3(tbl, protocol)
        self._gen_protocol_info_row4(tbl, protocol)

    # -------------------
    ## generate protocol info for row 0 of the header section
    # contains protocol id and description, start date
    #
    # @param tbl        the table to write to
    # @param protocol   the data to use
    # @return None
    def _gen_protocol_info_row0(self, tbl, protocol):
        row_cells = tbl.add_row().cells

        if self._do_results:
            start_date = protocol['start_date']
        else:
            start_date = ''

        self._gen_protocol_info_item(row_cells[0], row_cells[3], protocol['proto_id'], protocol['desc'])
        self._gen_protocol_info_item(row_cells[4], row_cells[6], 'Start date', start_date)

    # -------------------
    ## generate protocol info for row 1 of the header section
    # contains location of the protocol
    #
    # @param tbl        the table to write to
    # @param protocol   the data to use
    # @return None
    def _gen_protocol_info_row1(self, tbl, protocol):
        row_cells = tbl.add_row().cells

        who = f'{protocol["executed_by"]} at {protocol["location"]}'
        self._gen_protocol_info_item(row_cells[0], row_cells[3], 'At', who)
        self._gen_protocol_info_item(row_cells[4], row_cells[6], 'Requirements', 'see below')

    # -------------------
    ## generate protocol info for row 2 of the header section
    # contains softawre version and serial number of DUT
    #
    # @param tbl        the table to write to
    # @param protocol   the data to use
    # @return None
    def _gen_protocol_info_row2(self, tbl, protocol):
        row_cells = tbl.add_row().cells

        self._gen_protocol_info_item(row_cells[0], row_cells[3], 'Software Version', protocol['dut_version'])
        self._gen_protocol_info_item(row_cells[4], row_cells[6], 'Serial number', protocol['dut_serialno'])

    # -------------------
    ## generate protocol info for row 3 of the header section
    # contains Objectives of the protocol
    #
    # @param tbl        the table to write to
    # @param protocol   the data to use
    # @return None
    def _gen_protocol_info_row3(self, tbl, protocol):
        row_cells = tbl.add_row().cells
        self._gen_protocol_info_list(row_cells[0], row_cells[6], 'Objectives', protocol['objectives'])

    # -------------------
    ## generate protocol info for row 4 of the header section
    # contains Preconditions and Deviations if any
    #
    # @param tbl        the table to write to
    # @param protocol   the data to use
    # @return None
    def _gen_protocol_info_row4(self, tbl, protocol):
        row_cells = tbl.add_row().cells
        self._gen_protocol_info_list(row_cells[0], row_cells[3], 'Preconditions', protocol['preconditions'])
        self._gen_protocol_info_list(row_cells[4], row_cells[6], 'Deviations', protocol['deviations'])

    # -------------------
    ## generate protocol info for the given cell
    #
    # @param cell       to write to
    # @param last_cell  the last cell of the column span
    # @param title      the title of the cell
    # @param item       the data to use
    # @return None
    def _gen_protocol_info_item(self, cell, last_cell, title, item):
        self._cell_init(cell)
        self._cell_add_bold_text(cell, f'{title}: ')
        self._cell_add_text(cell, item)
        cell.merge(last_cell)
        self._set_cell_blue_background(cell)

    # -------------------
    ## generate protocol info for the given cell in list format
    #
    # @param cell       to write to
    # @param last_cell  the last cell of the column span
    # @param tag        the tag for the list
    # @param items      the list of items to use
    # @return None
    def _gen_protocol_info_list(self, cell, last_cell, tag, items):
        self._gen_list(cell, tag, items)
        cell.merge(last_cell)
        self._set_cell_blue_background(cell)

    # -------------------
    ## generate a list using the given items
    #
    # @param cell     the cell to write the list to
    # @param tag      the name of the list
    # @param items    the list of data to use
    # @return a paragraph containing the formatted list
    def _gen_list(self, cell, tag, items):
        self._cell_init(cell)
        self._cell_add_bold_text(cell, f'{tag}: ')

        if len(items) == 0:
            list_items = 'N/A'
            self._cell_add_text(cell, list_items)
        elif len(items) == 1:
            list_items = items[0]
            self._cell_add_text(cell, list_items)
        else:
            for item in items:
                self._cell_add_listitem(cell, item)

    # -------------------
    ## generate title row and one row per step
    #
    # @param tbl            the table to write to
    # @param steps          the steps data
    # @param requirements   the requirements found in all steps
    # @return None
    def _gen_steps_info(self, tbl, steps, requirements):
        # row 5: title row for step info
        row_cells = tbl.add_row().cells
        self._set_table_title(row_cells[0], 'Step')
        self._set_table_title(row_cells[1], 'Req.')
        self._set_table_title(row_cells[2], 'Execution Instructions')
        self._set_table_title(row_cells[3], 'Expected')
        self._set_table_title(row_cells[4], 'Actual')
        self._set_table_title(row_cells[5], 'Pass/\nFail')
        self._set_table_title(row_cells[6], 'Info')

        need_space = False
        if len(steps) == 0:
            row_cells = tbl.add_row().cells
            self._set_table_cell(row_cells[0], 'N/A')
            self._set_table_cell(row_cells[2], 'No steps found')
        else:
            # row N: one row for each step
            stepno = 0
            for step in steps:
                stepno += 1
                ns = self._gen_step(tbl, stepno, step, requirements)
                need_space = need_space or ns

        return need_space

    # -------------------
    ## generate a step in the test protocol
    #
    # @param tbl            the table to generate into
    # @param stepno         the step number of this step
    # @param step           the step description
    # @param requirements   the one or more reqids for this step
    # @return a paragraph containing the formatted list
    def _gen_step(self, tbl, stepno, step, requirements):
        # default is a passed, and empty result
        rs = ResultSummary()
        self._get_overall_result(step, rs)

        row_cells = tbl.add_row().cells
        self._set_table_cell(row_cells[0], self._get_stepno('docx', stepno))
        row_cells[0].paragraphs[0].alignment = self._align_center()
        self._set_table_cell(row_cells[1], self._get_reqids('docx', rs, requirements))
        self._set_table_cell(row_cells[2], self._get_desc('docx', step))
        self._set_table_cell_mono(row_cells[3], self._get_expected('docx', rs))
        self._set_table_cell_mono(row_cells[4], self._get_actual('docx', rs))
        self._set_table_cell(row_cells[5], self._get_result('docx', rs))
        row_cells[5].paragraphs[0].alignment = self._align_center()
        self._set_table_cell(row_cells[6], self._get_details('docx', rs, step))

        if self._do_results:
            self._gen_step_comments(tbl, step)

        # if the actual or expected is long-ish, indicate those columsn will need more space
        if services.cfg.page_info.orientation == 'portrait':
            max_len = 10
        else:
            max_len = 20
        if len(str(rs.actual_formatted)) > max_len:
            return True
        if len(str(rs.expected_formatted)) > max_len:
            return True

        return False

    # -------------------
    ## generate the comments for the given step
    #
    # @param tbl    the table to add the comments to
    # @param step   the step data
    # @return None
    def _gen_step_comments(self, tbl, step):
        for comment in step['comments']:
            para = f'Note: {comment}'
            row_cells = tbl.add_row().cells
            self._set_table_cell(row_cells[2], para)

    # -------------------
    ## generate a protocol table
    #
    # @return tbl object
    def _gen_protocol_table(self):
        tbl = self._doc.add_table(rows=0, cols=7)
        tbl.style = 'Table Grid'
        self._align_table_center(tbl)

        return tbl

    # -------------------
    ## update protocol table layout
    #
    # @param tbl          the table to write to
    # @param need_space   indicates if the table requires a different format because of space layout
    # @return None
    def _update_protocol_table(self, tbl, need_space):
        # these widths are the same portrait, landscape, needs space or not
        stepno_width = docx.shared.Inches(0.41)
        reqmt_width = docx.shared.Inches(0.65)
        details_width = docx.shared.Inches(1.47)
        result_width = docx.shared.Inches(0.47)
        if services.cfg.page_info.orientation == 'portrait':
            # portrait: (7.5" width)
            if need_space:
                value_width = docx.shared.Inches(1.31)
                desc_width = docx.shared.Inches(1.88)
            else:
                value_width = docx.shared.Inches(0.70)
                desc_width = docx.shared.Inches(3.10)
        else:
            # landscape (10" width)
            if need_space:
                value_width = docx.shared.Inches(1.68)
                desc_width = docx.shared.Inches(3.64)
            else:
                value_width = docx.shared.Inches(1.00)
                desc_width = docx.shared.Inches(5.00)

        tbl.columns[0].width = stepno_width  # stepno
        tbl.columns[1].width = reqmt_width  # reqmt.
        tbl.columns[2].width = desc_width  # desc
        tbl.columns[3].width = value_width  # expected
        tbl.columns[4].width = value_width  # actual
        tbl.columns[5].width = result_width  # result
        tbl.columns[6].width = details_width  # details

        # uncomment to debug
        # total = 0.0
        # units = 913780.48780487  # width units per inch
        # for i in range(0, 6 + 1):
        #     w = tbl.columns[i].width
        #     total += w
        #     logger.dbg(f"    {i} {w / units: >8.2f} docx.shared.Inches")
        # # there are 913780 per inch
        # logger.dbg(f"{total / units: >8.2f} docx.shared.Inches, need_space:{need_space}")

    # -------------------
    ## generate a requirement table
    #
    # @param requirements   a list of reqmt info
    # @return None
    def _gen_reqmt_table(self, requirements):
        # row 0: title row
        tbl = self._doc.add_table(rows=0, cols=2)
        tbl.style = 'Table Grid'
        self._align_table_center(tbl)
        #  = tbl._tblPr

        if services.cfg.page_info.orientation == 'portrait':
            # portrait: 7.5"
            tbl.columns[0].width = docx.shared.Inches(0.85)  # reqmt.
            tbl.columns[1].width = docx.shared.Inches(6.65)  # desc
        else:
            # landscape: 10"
            tbl.columns[0].width = docx.shared.Inches(0.85)  # reqmt.
            tbl.columns[1].width = docx.shared.Inches(9.15)  # desc

        row_cells = tbl.add_row().cells
        self._set_table_title(row_cells[0], 'Req.')
        self._set_table_title(row_cells[1], 'Desc.')

        if len(requirements) == 0:
            row_cells = tbl.add_row().cells
            self._set_table_cell(row_cells[0], 'N/A')
            self._set_table_cell(row_cells[1], 'No requirements found')
        else:
            for reqmt in sorted(requirements):
                row_cells = tbl.add_row().cells
                self._set_table_cell(row_cells[0], reqmt)

                if reqmt in self._requirements:
                    desc = self._requirements[reqmt]['desc']
                else:
                    desc = f'Could not find {reqmt} in file: {services.cfg.reqmt_json_path}'
                self._set_table_cell(row_cells[1], desc)

    # -------------------
    ## generate a signature area
    #
    # @return None
    def _gen_testers(self):
        self._gen_title('Manual Tester Signatures')

        # row 0: title row
        tbl = self._doc.add_table(rows=0, cols=3)
        tbl.style = 'Table Grid'
        self._align_table_center(tbl)

        if services.cfg.page_info.orientation == 'portrait':
            # portrait: 7.5"
            tbl.columns[0].width = docx.shared.Inches(0.50)  # initials
            tbl.columns[1].width = docx.shared.Inches(3.50)  # name
            tbl.columns[2].width = docx.shared.Inches(3.50)  # signature
        else:
            # landscape: 10"
            tbl.columns[0].width = docx.shared.Inches(0.50)  # reqmt.
            tbl.columns[1].width = docx.shared.Inches(3.25)  # name
            tbl.columns[2].width = docx.shared.Inches(6.25)  # signature

        row_cells = tbl.add_row().cells
        self._set_table_title(row_cells[0], '')
        self._set_table_title(row_cells[1], 'Tester')
        self._set_table_title(row_cells[2], 'Signature')

        for initials in sorted(self._testers):
            if initials == 'manual':
                initials = ''
                name = ''
            elif initials in services.cfg.testers:
                name = services.cfg.testers[initials]
            else:
                name = ''
            row_cells = tbl.add_row().cells
            self._set_table_cell(row_cells[0], initials)
            self._set_table_cell(row_cells[1], name)
            self._set_table_cell(row_cells[2], '\n\n')

    # -------------------
    ## set a table title to the given text
    #
    # @param cell   the table cell to write to
    # @param text   the title text to write
    # @return None
    def _set_table_title(self, cell, text):
        cell.text = ''
        cell.paragraphs[0].add_run(text).bold = True
        cell.paragraphs[0].style = 'ver_table1_header'
        self._set_cell_grey_background(cell)
        self._align_vertical_center(cell)

    # -------------------
    ## set a table cell to the given text
    #
    # @param cell   the table cell to write to
    # @param text   the text to write
    # @return None
    def _set_table_cell(self, cell, text):
        cell.text = ''
        cell.paragraphs[0].add_run(text)
        cell.paragraphs[0].style = 'ver_table1_cell'
        self._align_vertical_center(cell)

    # -------------------
    ## set a table cell to the given text in monospace font
    #
    # @param cell   the table cell to write to
    # @param text   the text to write
    # @return None
    def _set_table_cell_mono(self, cell, text):
        cell.text = ''
        cell.paragraphs[0].add_run(text)
        cell.paragraphs[0].style = 'ver_monospace'
        self._align_vertical_center(cell)

    # -------------------
    ## initialize a table cell
    #
    # @param cell   the table cell to write to
    # @return None
    def _cell_init(self, cell):
        cell.text = ''
        cell.paragraphs[0].style = 'ver_table1_cell'
        self._align_vertical_center(cell)

    # -------------------
    ## add bold text to the given table cell
    #
    # @param cell   the table cell to write to
    # @param text   the text to write
    # @return None
    def _cell_add_bold_text(self, cell, text):
        run = cell.paragraphs[0].add_run(text)
        run.bold = True

    # -------------------
    ## add text to the given table cell
    #
    # @param cell   the table cell to write to
    # @param text   the text to write
    # @return None
    def _cell_add_text(self, cell, text):
        cell.paragraphs[0].add_run(text)

    # -------------------
    ## add a list item the given table cell
    #
    # @param cell   the table cell to write to
    # @param item   the list item to write
    # @return None
    def _cell_add_listitem(self, cell, item):
        para = cell.add_paragraph(style='ver_list_item')
        para.add_run(item)

    # -------------------
    ## build the document with the current list of elements
    #
    # @return None
    def _build(self):
        self._doc.save(self._doc_path)



from .footer import Footer
from .header import Header
from medver_pytest import services


# -------------------
## Holds common page details
class PageDetails:
    # -------------------
    ## constuctor
    def __init__(self):
        ## holds the page orientation: portait or landscape
        self.orientation = 'portait'

        ## holds the page size: letter or A4
        self.page_size = 'letter'

        ## holds header info
        self.header = Header()

        ## holds footer info
        self.footer = Footer()

    # -------------------
    ## save a header item
    #
    # @param item    the name of the item to set: left, middle or right
    # @param value   the value to set
    # @return None
    def __setitem__(self, item, value):
        if item == 'orientation':
            self.orientation = value
            return

        if item == 'page_size':
            self.page_size = value
            return

        services.abort(f'bad item name: {item}')

    # -------------------
    ## get a value from this object
    #
    # @param item   the name of the attribute to get
    # @return the value of the attribute
    def __getitem__(self, item):
        if item == 'header':
            return self.header

        if item == 'footer':
            return self.footer

        if item == 'orientation':
            return self.orientation

        if item == 'page_size':
            return self.page_size

        services.abort(f'bad item name: {item}')
        return None  # for lint



import sys


# -------------------
## Holds all info for logging debug lines
class StdoutLogger:
    ## flag to log to stdout or not
    verbose = True

    # --------------------
    ## log a message. Use ok() or err() as appropriate.
    #
    # @param ok      the check state
    # @param msg     the message to print
    # @param prefix  (optional) prefix for each line printed
    # @return None
    @staticmethod
    def log(ok, msg, prefix=None):
        if ok:
            StdoutLogger.ok(msg, prefix)
        else:
            StdoutLogger.err(msg, prefix)

    # --------------------
    ## log a series of messages. Use ok() or err() as appropriate.
    #
    # @param ok      the check state
    # @param title   the line indicating what the check is about
    # @param msgs    individual list of lines to print
    # @param prefix  (optional) prefix for each line printed
    # @return None
    @staticmethod
    def log_all(ok, title, msgs, prefix=None):
        StdoutLogger.log(ok, f'{title}: {ok}', prefix)
        for msg in msgs:
            StdoutLogger.log(ok, f'   - {msg}', prefix)

    # -------------------
    ## write a "====" line with the given message
    #
    # @param msg     the message to write
    # @param prefix  (optional) prefix for each line printed
    # @return None
    @staticmethod
    def start(msg, prefix=None):
        StdoutLogger._write_line('====', msg, prefix)

    # -------------------
    ## write a "line" line with the given message
    #
    # @param msg     the message to write
    # @param prefix  (optional) prefix for each line printed
    # @return None
    @staticmethod
    def line(msg, prefix=None):
        StdoutLogger._write_line(' ', msg, prefix)

    # -------------------
    ## write a "ok" line with the given message
    #
    # @param msg     the message to write
    # @param prefix  (optional) prefix for each line printed
    # @return None
    @staticmethod
    def ok(msg, prefix=None):
        StdoutLogger._write_line('OK', msg, prefix)

    # -------------------
    ## write a "err" line with the given message
    #
    # @param msg     the message to write
    # @param prefix  (optional) prefix for each line printed
    # @return None
    @staticmethod
    def err(msg, prefix=None):
        StdoutLogger._write_line('ERR', msg, prefix)

    # -------------------
    ## write a "bug" line with the given message
    #
    # @param msg     the message to write
    # @param prefix  (optional) prefix for each line printed
    # @return None
    @staticmethod
    def bug(msg, prefix=None):
        StdoutLogger._write_line('BUG', msg, prefix)

    # -------------------
    ## write an output line with the given message
    #
    # @param msg     the message to write
    # @param prefix  (optional) prefix for each line printed
    # @return None
    @staticmethod
    def output(msg, prefix=None):
        StdoutLogger._write_line(' -- ', msg, prefix)

    # -------------------
    ## write a "warn" line with the given message
    #
    # @param msg     the message to write
    # @param prefix  (optional) prefix for each line printed
    # @return None
    @staticmethod
    def warn(msg, prefix=None):
        StdoutLogger._write_line('WARN', msg, prefix)

    # -------------------
    ## write a "err" line with the given message
    #
    # @param msg     the message to write
    # @param prefix  (optional) prefix for each line printed
    # @return None
    @staticmethod
    def dbg(msg, prefix=None):
        StdoutLogger._write_line('DBG', msg, prefix)

    # -------------------
    ## write the given line to stdout
    #
    # @param tag     the prefix tag
    # @param msg     the message to write
    # @param prefix  (optional) prefix for line
    # @return None
    @staticmethod
    def _write_line(tag, msg, prefix=None):
        if not StdoutLogger.verbose:
            return

        if prefix:
            print(f'{prefix} {tag: <4} {msg}')  # print okay
        else:
            print(f'{tag: <4} {msg}')  # print okay
        sys.stdout.flush()



from .gen_tp_base import GenTpBase
from ..common.gen_base_txt import GenBaseTxt
from medver_pytest import services
from medver_pytest.result_summary import ResultSummary


# -------------------
## Generates a Test Protocol report in text format
class GenTpTxt(GenBaseTxt, GenTpBase):
    # -------------------
    ## constructor
    #
    # @param protocols   the data to use
    # @param do_results  generate results or not (default True)
    def __init__(self, protocols, do_results=True):
        ## see base clase for doc
        self._doc_path = None
        ## see base clase for doc
        self._testers = None

        GenBaseTxt.__init__(self)
        GenTpBase.__init__(self, protocols, do_results)

        self._init('txt', protocols, do_results)

    # -------------------
    ## generate the report
    #
    # @return None
    def gen(self):
        self._gen_init()
        # uncomment to debug:
        # from medver_pytest.lib.stdout_logger import StdoutLogger as logger
        # logger.dbg(f'protocols: {json.dumps(self._protocols, indent=2)}')

        # TODO add data from self._requirements = Requirements.fast_load()

        with open(self._doc_path, 'w', encoding='UTF-8') as self._fp:
            self._gen_test_run_details()
            title = self._get_title()
            self._gen_title(title)

            for _, protocol in self._protocols.items():
                self._gen_protocol(protocol)

    # -------------------
    ## generate a protocol table using the given data
    #
    # @param protocol   the data to use
    # @return None
    def _gen_protocol(self, protocol):
        self._gen_protocol_info(protocol)

        requirements = []
        self._gen_steps_info(protocol['steps'], requirements)
        self._fp.write('\n')
        self._gen_reqmt_table(requirements)
        if self._testers:
            self._fp.write('\n')
            self._gen_testers()
        self._fp.write('\n\n')

    # -------------------
    ## generate the overall protocol information
    #
    # @param protocol  the protocol data
    # @return None
    def _gen_protocol_info(self, protocol):
        self._fp.write(f'==== protocol: {protocol["proto_id"]} {protocol["desc"]}\n')

        # row 0
        if self._do_results:
            start_date = f'     {"Start date": <17}: {protocol["start_date"]}\n'
        else:
            start_date = f'     {"Start date": <17}:\n'
        self._fp.write(start_date)

        # row 1
        self._fp.write(f'     {"At": <17}: {protocol["location"]}\n')
        self._fp.write(f'     {"Requirements": <17}: see below\n')

        # row 2
        self._fp.write(f'     {"Software Version": <17}: {protocol["dut_version"]}\n')
        self._fp.write(f'     {"Serial number": <17}: {protocol["dut_serialno"]}\n')

        # row 3
        self._fp.write(f'     {"Objectives": <17}: {protocol["objectives"]}\n')

        # row 4
        self._fp.write(f'     {"Preconditions": <17}: {protocol["preconditions"]}\n')
        self._fp.write(f'     {"Deviations": <17}: {protocol["deviations"]}\n')

    # -------------------
    ## generate title row and one row per step
    #
    # @param steps          the steps data
    # @param requirements   the requirements found in all steps
    # @return None
    def _gen_steps_info(self, steps, requirements):
        stepno = 0
        if len(steps) == 0:
            self._fp.write('     No steps found\n')
        else:
            for step in steps:
                stepno += 1
                self._gen_step(stepno, step, requirements)

    # -------------------
    ## generate a step in the test protocol
    #
    # @param stepno         the step number of this step
    # @param step           the step description
    # @param requirements   the one or more reqids for this step
    # @return None
    def _gen_step(self, stepno, step, requirements):
        self._fp.write(f'     Step {self._get_stepno("txt", stepno): <3}: {self._get_desc("txt", step)}\n')

        rs = ResultSummary()
        self._get_overall_result(step, rs)
        self._fp.write(f'       > reqids       : {self._get_reqids("txt", rs, requirements)}\n')
        self._fp.write(f'       > result       : {self._get_result("txt", rs)}\n')
        self._fp.write(f'       > expected     : {self._get_expected("txt", rs)}\n')
        self._fp.write(f'       > expected raw : {rs.expected}\n')
        self._fp.write(f'       > actual       : {self._get_actual("txt", rs)}\n')
        self._fp.write(f'       > actual raw   : {rs.actual}\n')
        self._fp.write(f'       > details      : {self._get_details("txt", rs, step)}\n')

        if self._do_results:
            self._gen_step_comments(step)

    # -------------------
    ## generate the comments for the given step
    #
    # @param step   the step data
    # @return None
    def _gen_step_comments(self, step):
        for comment in step['comments']:
            self._fp.write(f'       > Note         : {comment}\n')

    # -------------------
    ## generate a requirement table
    #
    # @param requirements   a list of reqmt info
    # @return None
    def _gen_reqmt_table(self, requirements):
        # row 0: title row
        self._fp.write(f'     {"Req.": <8} {"Desc.": <60}\n')
        self._fp.write(f'     {"-" * 8} {"-" * 60}\n')

        if len(requirements) == 0:
            self._fp.write(f'     {"N/A": <8} {"No requirements found"}\n')
        else:
            for reqmt in sorted(requirements):
                if reqmt in self._requirements:
                    desc = self._requirements[reqmt]['desc']
                else:
                    desc = f'Could not find {reqmt} in file: {services.cfg.reqmt_json_path}'

                self._fp.write(f'     {reqmt: <8} {desc}\n')

    # -------------------
    ## generate a signature area for all testers
    #
    # @return None
    def _gen_testers(self):
        self._fp.write('     Manual Tester Signatures\n')
        self._fp.write(f'     {"-" * 105}\n\n')

        self._fp.write(f'     {"": <3} {"Tester": <40} {"Signature": <60}\n')
        self._fp.write(f'     {"-" * 3} {"-" * 40} {"-" * 60}\n')

        for initials in sorted(self._testers):
            if initials == 'manual':
                initials = ''
                name = ''
            elif initials in services.cfg.testers:
                name = services.cfg.testers[initials]
            else:
                name = ''
            self._fp.write(f'\n     {initials: <3} {name: <40} {"_" * 60}\n')



from pytest_check import check

from . import services
from .result_summary import ResultSummary
from .utils import Utils
from medver_pytest.lib.view import Gui


# -------------------
## Holds information about the current protocol
class Protocol:
    # -------------------
    ## constructor
    def __init__(self):
        ## holds the list of protocol info found in test cases
        self._protocols = {}
        ## holds the current protocol
        self._protocol = None
        ## holds the current step
        self._step = None
        ## holds the current test script name
        self._current_test_script = None
        ## holds if the output json file needs to be saved
        self._dirty = False

    # -------------------
    ## initialize
    #
    # @return None
    def init(self):
        self._protocol = None
        self._step = None

    # -------------------
    ## terminate
    #
    # @return None
    def term(self):
        self.save()

    # -------------------
    ## create new protocol
    #
    # @param proto_id    the protocol id
    # @param desc        the protocol description
    # @return None
    def protocol(self, proto_id: str, desc):
        self._check_test_change()

        dts = Utils.get_dts(use_cfg_fmt=True)
        self._protocol = {
            'proto_id': proto_id.upper(),
            'desc': desc,
            'location': Utils.get_location(levels=2),
            'executed_by': 'automated',
            'start_date': dts,
            'dut_version': 'N/A',
            'dut_serialno': 'N/A',
            'objectives': [],
            'preconditions': [],
            'deviations': [],
            'steps': [],
        }
        self._dirty = True

        if self._protocol['proto_id'] in self._protocols:
            prev_loc = self._protocols[self._protocol['proto_id']]['location']
            services.logger.err(f'{self._protocol["proto_id"]} protocol id is already in use:')
            services.logger.err(f'   location: {self._protocol["location"]}')
            services.logger.err(f'   previous: {prev_loc}')
            services.abort()
        else:
            self._protocols[self._protocol['proto_id']] = self._protocol
            self._dirty = True

        self._step = None
        self.save()

    # -------------------
    ## get the current protocol id
    #
    ## @return if set the current protocol id, otherwise "none"
    @property
    def protocol_id(self):
        if self._protocol:
            return self._protocol['proto_id']

        return 'none'

    # -------------------
    ## check if the test script name has changed,
    # if so, save the current content and re-init protocols info
    #
    # @return None
    def _check_test_change(self):
        if self._current_test_script != services.cfg.test_script:
            if self._current_test_script is not None:
                self.save()
            self._current_test_script = services.cfg.test_script
            self._protocols = {}
            self._dirty = True

            services.trace.check_test_script()
            services.summary.check_test_script()

    # -------------------
    ## set the version for the Device Under Test (DUT)
    #
    # @param version     the version value
    # @return None
    def set_dut_version(self, version):
        self._check_test_change()

        if self._protocol is None:
            services.abort('set_dut_version(): missing pth.proto.protocol(), exiting')

        self._protocol['dut_version'] = version
        self._dirty = True

    # -------------------
    ## set the serial number for the Device Under Test (DUT)
    #
    # @param serialno     the serial number value
    # @return None
    def set_dut_serialno(self, serialno):
        self._check_test_change()

        if self._protocol is None:
            services.abort('set_dut_serialno(): missing pth.proto.protocol(), exiting')

        self._protocol['dut_serialno'] = serialno
        self._dirty = True

    # -------------------
    ## an objective of the current test case
    #
    # @param desc     the description
    # @return None
    def add_objective(self, desc):
        self._check_test_change()

        if self._protocol is None:
            services.abort('add_objective(): missing pth.proto.protocol(), exiting')

        self._protocol['objectives'].append(desc)
        self._dirty = True

    # -------------------
    ## a precondition of the current test case
    #
    # @param desc     the description
    # @return None
    def add_precondition(self, desc):
        self._check_test_change()

        if self._protocol is None:
            services.abort('add_precondition(): missing pth.proto.protocol(), exiting')

        self._protocol['preconditions'].append(desc)
        self._dirty = True

    # -------------------
    ## a deviation of the current test case
    #
    # @param desc     the description
    # @return None
    def add_deviation(self, desc):
        self._check_test_change()

        if self._protocol is None:
            services.abort('add_deviation(): missing pth.proto.protocol(), exiting')

        self._protocol['deviations'].append(desc)
        self._dirty = True

    # -------------------
    ## create a new step in the current test protocol
    #
    # @param desc     the description
    # @return None
    def step(self, desc):
        self._check_test_change()

        if self._protocol is None:
            services.abort('step(): missing pth.proto.protocol(), exiting')

        self._create_step(desc)

    # -------------------
    ## update tester initials from the initials used from the last run
    #
    # @return None
    def update_tester_initials(self):
        self._step['tester'] = services.status.last_tester_initials

    # -------------------
    ## save tester initials to step info
    #
    # @param initials   the tester initials to use
    # @return None
    def _set_tester_initials(self, initials):
        self._step['tester'] = initials

    # -------------------
    ## create a new manual step in the current test protocol
    #
    # @param desc     the description of the action the tester should take
    # @return None
    def step_manual_action(self, desc):
        self._check_test_change()

        location = Utils.get_location(2)

        # create a results summary and pre-populate it
        rs = ResultSummary()
        rs.location = location
        rs.formatter = None
        rs.expected = 'action done'
        rs.expected_type = 'str'
        rs.expected_formatted = rs.expected

        if self._protocol is None:
            services.abort('step_manual_action: missing pth.proto.protocol(), exiting')

        self._create_step(desc)

        # default value, should always be replaced by tester initials
        self._set_tester_initials('manual')

        if not services.view:
            services.view = Gui()

        state, reason = services.view.show_step_action(self._protocol['proto_id'],
                                                       desc)
        if state == 'abort':
            msg = 'step_manual_action: tester pressed Abort'
            services.harness.ver.report_abort(msg, rs.location)
            services.abort(msg)

        self.update_tester_initials()

        if state == 'cantdo':
            services.logger.err(f'{self._protocol["proto_id"]} manual step FAIL:')
            services.logger.err(f'   location: {self._protocol["location"]}')
            services.logger.err(f'   action:   {desc}')
            services.logger.err(f'   reason:   {reason}')

            self.comment(reason)
            rs.actual = 'action unsuccessful'
            rs.actual_type = 'str'
            rs.actual_formatted = rs.actual
            rs.failed()
            self.add_result(rs)

            # abort current pytest
            # TODO stop using pytest-check
            with check:  # pragma: no cover
                # coverage: iuvmode set during IUV and UT runs; never gets here
                check.set_no_tb()
                raise AssertionError(f'at {rs.location} action failed: {reason}')

            return

        if state != 'done':
            services.abort(f'step_manual_action: unknown state: {state}')

        rs.actual = 'action done'
        rs.actual_type = 'str'
        rs.actual_formatted = rs.actual
        rs.passed()
        self.add_result(rs)

    # -------------------
    ## create common step
    #
    # @param desc the dstep description
    # @return None
    def _create_step(self, desc):
        dts = Utils.get_dts(use_cfg_fmt=False)
        self._step = {
            'desc': desc,
            'dts': dts,
            'tester': 'auto',
            'comments': [],
            'results': [],
        }
        self._protocol['steps'].append(self._step)
        self._dirty = True
        self.save()

    # -------------------
    ## add a comment
    #
    # @param msg     the comment to add
    # @return None
    def comment(self, msg):
        self._check_test_change()

        if self._protocol is None:
            services.abort('comment(): missing pth.proto.protocol(), exiting')

        if self._step is None:
            services.abort('comment(): missing pth.proto.step(), exiting')

        self._step['comments'].append(msg)
        self._dirty = True

    # -------------------
    ## add a result to the current step
    # save the data to storage
    #
    # @param rs     the result summary to associate with this step
    # @return None
    def add_result(self, rs):
        if self._protocol is None:
            services.abort('add_result(): missing pth.proto.protocol(), exiting')

        if self._step is None:
            services.abort('add_result(): missing pth.proto.step(), exiting')

        self._step['results'].append(rs)
        self._dirty = True

        if rs.reqids is not None:
            services.trace.add_proto(rs.reqids, self._protocol['proto_id'], rs.location)

        services.summary.add_result(rs.reqids, self._protocol['proto_id'], rs.result)
        self.save()

    # -------------------
    ## save protocol info to storage
    #
    # @return None
    def save(self):
        services.storage.save_protocol(self._protocols)
        self._dirty = False

    # === for IUV only

    # -------------------
    ## (IUV only) get current protocol info
    #
    # @return protocol object
    def iuv_get_protocol(self):
        return self._protocol

    # -------------------
    ## (IUV only) get current dirty flag
    #
    # @return dirty flag
    def iuv_get_dirty(self):
        return self._dirty



from medver_pytest import services
from medver_pytest.requirements import Requirements


# -------------------
## Common functions for Summary reports
class GenSummaryBase:  # pylint: disable=too-few-public-methods
    # -------------------
    ## constructor
    #
    # @param summary  the summary info to use
    def __init__(self, summary):
        ## holds the summary info
        self._summary = summary
        ## holds reference to the Requirements object
        self._requirements = None

    # -------------------
    ## initialization
    #
    # @return None
    def _gen_init(self):
        services.logger.start('report: summary')

        self._requirements = Requirements()
        self._requirements.load()

    # -------------------
    ## calculate total number of requirements
    #
    # @return None
    def _get_total_requirements(self):
        total1 = self._requirements.total
        total2 = len(self._summary['reqids'])
        if total1 > total2:
            return total1

        return total2

    # -------------------
    ## count requirements: failing, passing, invalid
    #
    # @param total           total number of requirements
    # @param report_invalid  callback function that prints a message that a requirement id is invalid
    # @return tuple: failing, passing, invalid
    def _count_requirements(self, total, report_invalid):
        num_failing = 0
        num_passing = 0
        num_invalid = 0
        for reqid, item in self._summary['reqids'].items():
            if item['result'] == 'FAIL':
                num_failing += 1
            else:
                num_passing += 1

            if self._requirements.is_found and not self._requirements.is_valid(reqid):
                num_invalid += 1
                report_invalid(reqid)

        num_missing = total - (num_passing + num_failing) + num_invalid
        return num_failing, num_passing, num_invalid, num_missing

    # -------------------
    ## count the passing/failing protocols
    #
    # @return tuple: failing, passing
    def _count_protocols(self):
        num_failing = 0
        num_passing = 0
        for _, item in self._summary['protoids'].items():
            if item['result'] == 'FAIL':
                num_failing += 1
            else:
                num_passing += 1

        return num_failing, num_passing

    # -------------------
    ## calculate the percentage count is of total
    #
    # @param count  the current count
    # @param total  the current total
    # @return formatted string with percent
    def _pct(self, count, total):
        if total == 0:
            pct = 0
        else:
            pct = round((count * 100.0) / total, 1)
        return f'{pct: >6}%'



import json
import os

from medver_pytest import services


# -------------------
## Holds content for IUV data as stored
# TODO handle other storage types
class IUVData:
    # -------------------
    ## constructor
    def __init__(self):
        pass

    # -------------------
    ## initialize
    #
    # @return None
    def init(self):
        pass

    # -------------------
    ## load protocol JSON file
    #
    # @return json object
    def load_protocol(self):
        # TODO get the path from a common object e.g. harness
        path = os.path.join('out', 'iuv0', f'{services.cfg.test_script}_protocol.json')
        with open(path, 'r', encoding='utf=8') as fp:
            j = json.load(fp)
        return j

    # -------------------
    ## load trace JSON file
    #
    # @return json object
    def load_trace(self):
        # TODO get the path from a common object e.g. harness
        path = os.path.join('out', 'iuv0', f'{services.cfg.test_script}_trace.json')
        with open(path, 'r', encoding='utf=8') as fp:
            j = json.load(fp)
        return j

    # -------------------
    ## load summary JSON file
    #
    # @return json object
    def load_summary(self):
        # TODO get the path from a common object e.g. harness
        path = os.path.join('out', 'iuv0', f'{services.cfg.test_script}_summary.json')
        with open(path, 'r', encoding='utf=8') as fp:
            j = json.load(fp)
        return j



import os

from medver_pytest import services
from medver_pytest.utils import Utils


# -------------------
## Holds all info for the logger file
class Logger:
    # -------------------
    ## constructor
    def __init__(self):
        ## the logger file pointer
        self._fp = None
        ## the current line count of the log file; used to determine when to flush
        self._line_count = 0
        ## the current DTS count; used to determine when to write the DTS to the log
        self._dts_count = 0
        ## the last DTS; used to show the elapsed time
        self._dts_last = None

    # -------------------
    ## intialize
    #
    # @param fname   the file name to write to; default medver_pytest.log
    # @return None
    def init(self, fname='medver_pytest.log'):
        path = os.path.join(services.cfg.outdir, fname)
        self._fp = open(path, 'w')

    # -------------------
    ## write a "start" line with the given message
    #
    # @param msg   the message to write
    # @return None
    def start(self, msg):
        self._write_line('====', msg, so=False)

    # -------------------
    ## write a "line" line with the given message
    #
    # @param msg   the message to write
    # @return None
    def line(self, msg):
        self._write_line(' ', msg, so=False)

    # -------------------
    ## write a "ok" line with the given message
    #
    # @param msg   the message to write
    # @return None
    def ok(self, msg):
        self._write_line('OK', msg, so=True)

    # -------------------
    ## write a "warn" line with the given message
    #
    # @param msg   the message to write
    # @return None
    def warn(self, msg):
        self._write_line('WARN', msg, so=True)

    # -------------------
    ## write a "err" line with the given message
    #
    # @param msg   the message to write
    # @return None
    def err(self, msg):
        self._write_line('ERR', msg, so=True)

    # -------------------
    ## write a "raw" line with the given message
    #
    # @param msg   the message to write
    # @return None
    def raw(self, msg):
        self._write_line('', msg, so=True, raw=True)

    # -------------------
    ## write a "user" line with the given message
    # this indicates an action taken by a tester e.g. clicking, ctrl-c
    #
    # @param msg   the message to write
    # @return None
    def user(self, msg):
        self._write_line('USER', msg, so=False, raw=False)

    # -------------------
    ## write the given line to the log. Optionally write it stdout
    #
    # @param tag   the prefix tag
    # @param msg   the message to write
    # @param so    if True, write to stdout
    # @param raw   if True, write the line without the tag or timestemp
    # @return None
    def _write_line(self, tag, msg, so=False, raw=False):
        if self._dts_count == 0:
            # capture the current dts in UTC and save it for doing the delta per line
            self._dts_last = Utils.get_utc()
            dts = Utils.get_dts(use_cfg_fmt=True, use_time=self._dts_last)

            line = f'{"----": <4} {dts}'
            self._fp.write(f'{line}\n')

            # print full dts every so often
            self._dts_count = 100

        if raw:
            line = msg
            self._fp.write(msg)
        else:
            line = f'{tag: <4} {msg}'

            self._dts_count -= 1

            # add time delta to log lines; self_dts.last is always UTC
            diff = Utils.get_utc() - self._dts_last
            self._fp.write(f'{diff} {line}\n')

        # ok to flush?
        self._line_count += 1
        if self._line_count > 5:
            self._fp.flush()

        # ok to write to stdout?
        if services.cfg.iuvmode:  # pragma: no cover
            # coverage: iuvmode is only set during IUV and UT runs
            services.harness.iuv.mock_logger(line)
        elif so:  # pragma: no cover
            # coverage: not used during IUV and UT runs
            print(line)  # print okay



# -------------------
import sys


# -------------------
## Holds all info for logging to stdout.
# Normally used only when the logger is not avaialble or before it is configured
class LoggerStdout:
    # -------------------
    ## initialize
    #
    # @param fname   ignored; matches signature in Logger
    # @return None
    def init(self, fname='ignored'):
        pass

    # -------------------
    ## write a "start" line with the given message
    #
    # @param msg   the message to write
    # @return None
    def start(self, msg):
        self._write_line('====', msg)

    # -------------------
    ## write a "line" line with the given message
    #
    # @param msg   the message to write
    # @return None
    def line(self, msg):
        self._write_line(' ', msg)

    # -------------------
    ## write a "ok" line with the given message
    #
    # @param msg   the message to write
    # @return None
    def ok(self, msg):
        self._write_line('OK', msg)

    # -------------------
    ## write a "warn" line with the given message
    #
    # @param msg   the message to write
    # @return None
    def warn(self, msg):
        self._write_line('WARN', msg)

    # -------------------
    ## write a "err" line with the given message
    #
    # @param msg   the message to write
    # @return None
    def err(self, msg):
        self._write_line('ERR', msg)

    # -------------------
    ## write a "raw" line with the given message
    #
    # @param msg   the message to write
    # @return None
    def raw(self, msg):
        self._write_line('', msg)

    # -------------------
    ## write a "user" line with the given message
    # this indicates an action taken by a tester e.g. clicking, ctrl-c
    #
    # @param msg   the message to write
    # @return None
    def user(self, msg):
        self._write_line('USER', msg)

    # -------------------
    ## write the given line to stdout
    #
    # @param tag   the prefix tag
    # @param msg   the message to write
    # @return None
    def _write_line(self, tag, msg):
        print(f'{tag: <4} {msg}')  # print okay
        sys.stdout.flush()



import os

from .gen_trace_matrix_base import GenTraceMatrixBase
from ..common.gen_base_minhtml import GenBaseMinHtml
from medver_pytest import services
from medver_pytest.constants import Constants


# -------------------
## Generates a Trace Matrix report in min-html format
class GenTraceMatrixMinHtml(GenBaseMinHtml, GenTraceMatrixBase):
    # -------------------
    ## constructor
    #
    # @param matrix   the data to use
    def __init__(self, matrix):
        GenBaseMinHtml.__init__(self)
        GenTraceMatrixBase.__init__(self, matrix)

        ## holds path to the output file
        self._path = os.path.join(services.cfg.outdir, f'{Constants.trace_fname}-min.html')

    # -------------------
    ## generate the report
    #
    # @return None
    def gen(self):
        self._gen_init()

        with open(self._path, 'w', encoding='UTF-8') as self._fp:
            self._fp.write('<header id="ver-trace-title">\n')
            self._fp.write('    <h2 class="title">Trace Matrix</h2>\n')
            self._fp.write('</header>\n')
            self._fp.write('\n')

            self._gen_test_run_details()
            self._fp.write('\n')

            self._gen_title('Trace Matrix', arg_id='Matrix')
            self._load_matrix_requirements()
            self._gen_table_open(css_class='ver-trace-table')
            self._gen_thead_open()

            self._gen_tr_open(css_class='ver-trace-hdg1')
            self._gen_th('Req. Id', css_class='ver-trace-hdg-col1')
            self._gen_th('Protocol', css_class='ver-trace-hdg-col2')
            self._gen_tr_close()
            self._gen_thead_close()
            self._gen_tbody_open()
            ## see base class doc
            self._report_requirements(self._report_header, self._report_desc, self._report_details)
            self._gen_tbody_close()
            self._gen_table_close()

    # -------------------
    ## callback to report the current requirement and tester info
    #
    # @param req_id   the requirement id to report
    # @param info     the info to report
    # @return None
    def _report_header(self, req_id, info):
        # col1 of row
        self._fp.write('    <tr>\n')
        self._fp.write(f'        <td class="ver-trace-col1">{req_id}<br/>{info}</td>\n')

    # -------------------
    ## callback to report the current requirement's description (if available)
    #
    # @param desc     the description to report
    # @return None
    def _report_desc(self, desc):
        # col2, line1 of row
        self._fp.write('        <td class="ver-trace-col2">\n')
        self._fp.write(f'            <strong>Desc: </strong><em>{desc}</em><br/><br/>\n')

    # -------------------
    ## callback to report protocol id and info for the current requirement
    #
    # @param proto_id     the protocol id to report
    # @param proto_info   the protocol info to report
    # @return None
    def _report_details(self, proto_id, proto_info):
        # col2, line2 of row
        if proto_id is None and proto_info is None:
            self._fp.write('        </td>\n')
            self._fp.write('    </tr>\n')
        elif proto_id is None:
            self._fp.write(f'            <div>{proto_info}</div>\n')
        else:
            self._fp.write(f'            <div>{proto_id} {proto_info}</div>\n')



import json
import os

import jsmin

from medver_pytest import services


# -------------------
## holds content of the requirements json file
class Requirements:
    # -------------------
    ## constructor
    def __init__(self):
        ## holds the requirements info from the json file, if any
        self._requirements = None
        ## indicates if the json file was found
        self._is_found = False
        ## indicates the total number of requirements
        self._total = 0
        ## indicates the number of requirements that are marked "auto"
        self._total_automated = 0
        ## indicates the number of requirements that are marked "manual"
        self._total_manual = 0
        ## indicates the number of requirements that are marked "both"
        self._total_both = 0

    # -------------------
    ## return whether the json file was found or not
    #
    # @return True if found, False otherwise
    @property
    def is_found(self):
        return self._is_found

    # -------------------
    ## total number of requirements in the file
    #
    # @return number of requirements
    @property
    def total(self):
        return self._total

    # -------------------
    ## total number of requirements marked "auto"
    #
    # @return number of automated requirements
    @property
    def total_automated(self):
        return self._total_automated

    # -------------------
    ## total number of requirements marked "manual"
    #
    # @return number of manual requirements
    @property
    def total_manual(self):
        return self._total_manual

    # -------------------
    ## total number of requirements marked "both"
    #
    # @return number of automated/manual requirements
    @property
    def total_both(self):
        return self._total_both

    # -------------------
    ## checks if the given requirement id is in the requirements list
    #
    # @param reqid  the id to check
    # @return True if the requirement is in the file
    def is_valid(self, reqid):
        return reqid in self._requirements

    # -------------------
    ## iterate through all requirements with a callback with reqid and reqmt info
    #
    # @return None
    def all_reqids(self):
        for reqid, reqmt in sorted(self._requirements.items()):
            yield reqid, reqmt

    # -------------------
    ## add a missing requirement, i.e. named in a protocol but not in the reqmt json file
    #
    # @param req_id  the requiremt id to add
    # @return None
    def add_reqid(self, req_id):
        if req_id in self._requirements:
            return

        if self._is_found:
            self._requirements[req_id] = {
                'tester': 'invalid',
                'desc': f'{req_id} not found in {services.cfg.reqmt_json_path}'
            }
        else:
            self._requirements[req_id] = {
                'tester': 'auto',
                'desc': 'N/A'
            }

    # -------------------
    ## load the json file
    #
    # @return None
    def load(self):
        if services.cfg.reqmt_json_path is None:
            self._requirements = {}
            return

        if not os.path.isfile(services.cfg.reqmt_json_path):
            services.logger.err(f'requirement json file not found: {services.cfg.reqmt_json_path}')
            self._requirements = {}
            return

        with open(services.cfg.reqmt_json_path, 'r', encoding='utf-8') as fp:
            cleanj = jsmin.jsmin(fp.read())
            self._requirements = json.loads(cleanj)

        self._is_found = True

        self._total = 0
        self._total_automated = 0
        self._total_manual = 0
        self._total_both = 0
        for reqid, reqmt in self._requirements.items():
            if reqmt['tester'] == 'auto':
                self._total += 1
                self._total_automated += 1
            elif reqmt['tester'] == 'manual':
                self._total += 1
                self._total_manual += 1
            elif reqmt['tester'] == 'both':
                self._total += 1
                self._total_manual += 1
                self._total_automated += 1
                self._total_both += 1
            else:
                services.logger.err(f'{reqid}: found invalid "tester" value "{reqmt["tester"]}"')
                services.abort(f'invalid {services.cfg.reqmt_json_path}')

    # -------------------
    ## load the requirements json file
    #
    # @return None if not found or not set, otherwise the JSON content
    @staticmethod
    def fast_load():
        if services.cfg.reqmt_json_path is None:
            return {}

        if not os.path.isfile(services.cfg.reqmt_json_path):
            return {}

        with open(services.cfg.reqmt_json_path, 'r', encoding='utf-8') as fp:
            cleanj = jsmin.jsmin(fp.read())
            return json.loads(cleanj)
